import tensorflow

from PrognosAIs.Model.Architectures.Architecture import ClassificationNetworkArchitecture
from tensorflow.keras.layers import Add
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import Conv3D
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import Dropout
from tensorflow.keras.layers import GlobalAveragePooling2D
from tensorflow.keras.layers import GlobalAveragePooling3D
from tensorflow.keras.layers import Input
from tensorflow.keras.layers import MaxPooling2D
from tensorflow.keras.layers import MaxPooling3D
from tensorflow.keras.layers import ReLU
from tensorflow.keras.models import Model


class ResNet(ClassificationNetworkArchitecture):
    def get_residual_conv_block(
        self, layer: tensorflow.keras.layers, N_filters: int, kernel_size: list
    ):
        N_dims = len(kernel_size)

        if N_dims == 2:
            conv_func = Conv2D
        elif N_dims == 3:
            conv_func = Conv3D

        stride_size = self.get_corrected_stride_size(layer, [2] * N_dims, kernel_size)

        conv_1 = conv_func(
            filters=N_filters, kernel_size=kernel_size, strides=stride_size, padding="same",
        )(layer)
        batchnorm_1 = BatchNormalization(axis=-1)(conv_1)
        relu_1 = ReLU()(batchnorm_1)

        conv_2 = conv_func(
            filters=N_filters, kernel_size=kernel_size, strides=[1] * N_dims, padding="same"
        )(relu_1)
        batchnorm_2 = BatchNormalization(axis=-1)(conv_2)

        conv_skip = conv_func(
            filters=N_filters, kernel_size=[1] * N_dims, strides=stride_size, padding="valid"
        )(layer)
        batch_skip = BatchNormalization(axis=-1)(conv_skip)

        res_output = Add()([batch_skip, batchnorm_2])

        relu_output = ReLU()(res_output)

        return relu_output

    def get_residual_identity_block(self, layer, N_filters, kernel_size):
        N_dims = len(kernel_size)
        if N_dims == 2:
            conv_func = Conv2D
        elif N_dims == 3:
            conv_func = Conv3D

        conv_1 = conv_func(
            filters=N_filters, kernel_size=kernel_size, strides=[1] * N_dims, padding="same"
        )(layer)
        batchnorm_1 = BatchNormalization(axis=-1)(conv_1)
        relu_1 = ReLU()(batchnorm_1)

        conv_2 = conv_func(
            filters=N_filters, kernel_size=kernel_size, strides=[1] * N_dims, padding="same"
        )(relu_1)
        batchnorm_2 = BatchNormalization(axis=-1)(conv_2)

        res_output = Add()([batchnorm_2, layer])
        relu_output = ReLU()(res_output)

        return relu_output


class ResNet_18_2D(ResNet):
    def create_model(self):
        self.check_minimum_input_size(self.inputs, [9, 9])
        stride_size = self.get_corrected_stride_size(self.inputs, [2, 2], [7, 7])

        conv_1 = Conv2D(
            filters=64, kernel_size=[7, 7], strides=stride_size, padding="valid", activation="relu",
        )(self.inputs)

        stride_size = self.get_corrected_stride_size(conv_1, [2, 2], [3, 3])

        pooling_1 = MaxPooling2D(pool_size=[3, 3], strides=stride_size, padding="valid",)(conv_1)

        res_block_1 = self.get_residual_identity_block(pooling_1, 64, [3, 3])
        res_block_2 = self.get_residual_identity_block(res_block_1, 64, [3, 3])

        res_block_2 = self.make_dropout_layer(res_block_2)

        res_block_3 = self.get_residual_conv_block(res_block_2, 128, [3, 3])
        res_block_4 = self.get_residual_identity_block(res_block_3, 128, [3, 3])

        res_block_4 = self.make_dropout_layer(res_block_4)

        res_block_5 = self.get_residual_conv_block(res_block_4, 256, [3, 3])
        res_block_6 = self.get_residual_identity_block(res_block_5, 256, [3, 3])

        res_block_6 = self.make_dropout_layer(res_block_6)

        res_block_7 = self.get_residual_conv_block(res_block_6, 512, [3, 3])
        res_block_8 = self.get_residual_identity_block(res_block_7, 512, [3, 3])

        gap_1 = GlobalAveragePooling2D()(res_block_8)

        gap_1 = self.make_dropout_layer(gap_1)

        predictions = self.outputs(gap_1)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model


class ResNet_34_2D(ResNet):
    def create_model(self):
        self.check_minimum_input_size(self.inputs, [9, 9])
        stride_size = self.get_corrected_stride_size(self.inputs, [2, 2], [7, 7])

        conv_1 = Conv2D(
            filters=64, kernel_size=[7, 7], strides=stride_size, padding="valid", activation="relu",
        )(self.inputs)

        stride_size = self.get_corrected_stride_size(conv_1, [2, 2], [3, 3])

        pooling_1 = MaxPooling2D(pool_size=[3, 3], strides=stride_size, padding="valid",)(conv_1)

        res_block_1 = self.get_residual_identity_block(pooling_1, 64, [3, 3])
        res_block_2 = self.get_residual_identity_block(res_block_1, 64, [3, 3])
        res_block_3 = self.get_residual_identity_block(res_block_2, 64, [3, 3])

        res_block_3 = self.make_dropout_layer(res_block_3)

        res_block_4 = self.get_residual_conv_block(res_block_3, 128, [3, 3])
        res_block_5 = self.get_residual_identity_block(res_block_4, 128, [3, 3])
        res_block_6 = self.get_residual_identity_block(res_block_5, 128, [3, 3])
        res_block_7 = self.get_residual_identity_block(res_block_6, 128, [3, 3])

        res_block_7 = self.make_dropout_layer(res_block_7)

        res_block_8 = self.get_residual_conv_block(res_block_7, 256, [3, 3])
        res_block_9 = self.get_residual_identity_block(res_block_8, 256, [3, 3])
        res_block_10 = self.get_residual_identity_block(res_block_9, 256, [3, 3])
        res_block_11 = self.get_residual_identity_block(res_block_10, 256, [3, 3])
        res_block_12 = self.get_residual_identity_block(res_block_11, 256, [3, 3])
        res_block_13 = self.get_residual_identity_block(res_block_12, 256, [3, 3])

        res_block_13 = self.make_dropout_layer(res_block_13)

        res_block_14 = self.get_residual_conv_block(res_block_13, 512, [3, 3])
        res_block_15 = self.get_residual_identity_block(res_block_14, 512, [3, 3])
        res_block_16 = self.get_residual_identity_block(res_block_15, 512, [3, 3])

        gap_1 = GlobalAveragePooling2D()(res_block_16)

        gap_1 = self.make_dropout_layer(gap_1)

        predictions = self.outputs(gap_1)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model


class ResNet_18_3D(ResNet):
    def create_model(self):
        self.check_minimum_input_size(self.inputs, [9, 9, 9])

        stride_size = self.get_corrected_stride_size(self.inputs, [2, 2, 2], [7, 7, 7])

        conv_1 = Conv3D(
            filters=64,
            kernel_size=[7, 7, 7],
            strides=stride_size,
            padding="valid",
            activation="relu",
        )(self.inputs)

        stride_size = self.get_corrected_stride_size(conv_1, [2, 2, 2], [3, 3, 3])

        pooling_1 = MaxPooling3D(pool_size=[3, 3, 3], strides=stride_size, padding="valid",)(conv_1)

        res_block_1 = self.get_residual_identity_block(pooling_1, 64, [3, 3, 3])
        res_block_2 = self.get_residual_identity_block(res_block_1, 64, [3, 3, 3])

        res_block_2 = self.make_dropout_layer(res_block_2)

        res_block_3 = self.get_residual_conv_block(res_block_2, 128, [3, 3, 3])
        res_block_4 = self.get_residual_identity_block(res_block_3, 128, [3, 3, 3])

        res_block_4 = self.make_dropout_layer(res_block_4)

        res_block_5 = self.get_residual_conv_block(res_block_4, 256, [3, 3, 3])
        res_block_6 = self.get_residual_identity_block(res_block_5, 256, [3, 3, 3])

        res_block_6 = self.make_dropout_layer(res_block_6)

        res_block_7 = self.get_residual_conv_block(res_block_6, 512, [3, 3, 3])
        res_block_8 = self.get_residual_identity_block(res_block_7, 512, [3, 3, 3])

        gap_1 = GlobalAveragePooling3D()(res_block_8)

        gap_1 = self.make_dropout_layer(gap_1)

        predictions = self.outputs(gap_1)

        model = Model(inputs=self.inputs, outputs=predictions)
        return model


class ResNet_18_multioutput_3D(ResNet):
    def create_model(self):
        self.check_minimum_input_size(self.inputs, [9, 9, 9])

        stride_size = self.get_corrected_stride_size(self.inputs, [2, 2, 2], [7, 7, 7])

        conv_1 = Conv3D(
            filters=64,
            kernel_size=[7, 7, 7],
            strides=stride_size,
            padding="valid",
            activation="relu",
        )(self.inputs)

        stride_size = self.get_corrected_stride_size(conv_1, [2, 2, 2], [3, 3, 3])

        pooling_1 = MaxPooling3D(pool_size=[3, 3, 3], strides=stride_size, padding="valid",)(conv_1)

        res_block_1 = self.get_residual_identity_block(pooling_1, 64, [3, 3, 3])
        res_block_2 = self.get_residual_identity_block(res_block_1, 64, [3, 3, 3])

        res_block_2 = self.make_dropout_layer(res_block_2)

        res_block_3 = self.get_residual_conv_block(res_block_2, 128, [3, 3, 3])
        res_block_4 = self.get_residual_identity_block(res_block_3, 128, [3, 3, 3])

        res_block_4 = self.make_dropout_layer(res_block_4)

        res_block_5 = self.get_residual_conv_block(res_block_4, 256, [3, 3, 3])
        res_block_6 = self.get_residual_identity_block(res_block_5, 256, [3, 3, 3])

        res_block_6 = self.make_dropout_layer(res_block_6)

        predictions = []
        for i_output in self.outputs.values():
            res_block_7_output_branch = self.get_residual_conv_block(res_block_6, 512, [3, 3, 3])
            res_block_8_output_branch = self.get_residual_identity_block(
                res_block_7_output_branch, 512, [3, 3, 3]
            )

            gap_1_output_branch = GlobalAveragePooling3D()(res_block_8_output_branch)

            gap_1_output_branch = self.make_dropout_layer(gap_1_output_branch)

            predictions.append(i_output(gap_1_output_branch))

        model = Model(inputs=self.inputs, outputs=predictions)
        return model


class ResNet_34_3D(ResNet):
    def create_model(self):
        self.check_minimum_input_size(self.inputs, [9, 9, 9])
        stride_size = self.get_corrected_stride_size(self.inputs, [2, 2, 2], [7, 7, 7])

        conv_1 = Conv3D(
            filters=64,
            kernel_size=[7, 7, 7],
            strides=stride_size,
            padding="valid",
            activation="relu",
        )(self.inputs)

        stride_size = self.get_corrected_stride_size(conv_1, [2, 2, 2], [3, 3, 3])

        pooling_1 = MaxPooling3D(pool_size=[3, 3, 3], strides=stride_size, padding="valid",)(conv_1)

        res_block_1 = self.get_residual_identity_block(pooling_1, 64, [3, 3, 3])
        res_block_2 = self.get_residual_identity_block(res_block_1, 64, [3, 3, 3])
        res_block_3 = self.get_residual_identity_block(res_block_2, 64, [3, 3, 3])

        res_block_3 = self.make_dropout_layer(res_block_3)

        res_block_4 = self.get_residual_conv_block(res_block_3, 128, [3, 3, 3])
        res_block_5 = self.get_residual_identity_block(res_block_4, 128, [3, 3, 3])
        res_block_6 = self.get_residual_identity_block(res_block_5, 128, [3, 3, 3])
        res_block_7 = self.get_residual_identity_block(res_block_6, 128, [3, 3, 3])

        res_block_7 = self.make_dropout_layer(res_block_7)

        res_block_8 = self.get_residual_conv_block(res_block_7, 256, [3, 3, 3])
        res_block_9 = self.get_residual_identity_block(res_block_8, 256, [3, 3, 3])
        res_block_10 = self.get_residual_identity_block(res_block_9, 256, [3, 3, 3])
        res_block_11 = self.get_residual_identity_block(res_block_10, 256, [3, 3, 3])
        res_block_12 = self.get_residual_identity_block(res_block_11, 256, [3, 3, 3])
        res_block_13 = self.get_residual_identity_block(res_block_12, 256, [3, 3, 3])

        res_block_13 = self.make_dropout_layer(res_block_13)

        res_block_14 = self.get_residual_conv_block(res_block_13, 512, [3, 3, 3])
        res_block_15 = self.get_residual_identity_block(res_block_14, 512, [3, 3, 3])
        res_block_16 = self.get_residual_identity_block(res_block_15, 512, [3, 3, 3])

        gap_1 = GlobalAveragePooling3D()(res_block_16)

        gap_1 = self.make_dropout_layer(gap_1)

        predictions = self.outputs(gap_1)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model



from PrognosAIs.Model.Architectures.Architecture import ClassificationNetworkArchitecture
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import Conv3D
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import Dropout
from tensorflow.keras.layers import Flatten
from tensorflow.keras.layers import MaxPooling2D
from tensorflow.keras.layers import MaxPooling3D
from tensorflow.keras.models import Model


class VGG(ClassificationNetworkArchitecture):
    def init_dimensionality(self, N_dimension):
        if N_dimension == 2:
            self.dims = 2
            self.conv_func = Conv2D
            self.pool_func = MaxPooling2D
        elif N_dimension == 3:
            self.dims = 3
            self.conv_func = Conv3D
            self.pool_func = MaxPooling3D

    def get_VGG_block(self, layer, N_filters, N_conv_layer):
        stride_size = self.get_corrected_stride_size(layer, [2] * self.dims, [2] * self.dims)

        pooling = self.pool_func(pool_size=[2] * self.dims, strides=stride_size, padding="valid")(
            layer
        )

        conv = self.conv_func(
            filters=N_filters,
            kernel_size=[3] * self.dims,
            strides=[1] * self.dims,
            padding="same",
            activation="relu",
        )(pooling)

        for i_conv_layer in range(N_conv_layer - 1):
            conv = self.conv_func(
                filters=N_filters,
                kernel_size=[3] * self.dims,
                strides=[1] * self.dims,
                padding="same",
                activation="relu",
            )(conv)

        return conv


class VGG_16_2D(VGG):
    dims = 2

    def create_model(self):
        self.init_dimensionality(self.dims)
        self.check_minimum_input_size(self.inputs, [24, 24])

        conv_1 = Conv2D(
            filters=64, kernel_size=[3, 3], strides=[1, 1], padding="same", activation="relu",
        )(self.inputs)

        conv_2 = Conv2D(
            filters=64, kernel_size=[3, 3], strides=[1, 1], padding="same", activation="relu",
        )(conv_1)

        vgg_block_1 = self.get_VGG_block(conv_2, 128, 2)
        vgg_block_1 = self.make_dropout_layer(vgg_block_1)
        vgg_block_2 = self.get_VGG_block(vgg_block_1, 256, 3)
        vgg_block_2 = self.make_dropout_layer(vgg_block_2)
        vgg_block_3 = self.get_VGG_block(vgg_block_2, 512, 3)
        vgg_block_3 = self.make_dropout_layer(vgg_block_3)
        vgg_block_4 = self.get_VGG_block(vgg_block_3, 512, 3)

        stride_size = self.get_corrected_stride_size(vgg_block_4, [2, 2], [2, 2])
        pooling_1 = MaxPooling2D(pool_size=[2, 2], strides=stride_size, padding="valid",)(
            vgg_block_4
        )
        flatten_1 = Flatten()(pooling_1)
        flatten_1 = self.make_dropout_layer(flatten_1)
        dense_1 = Dense(4096, activation="relu")(flatten_1)
        dense_2 = Dense(4096, activation="relu")(dense_1)
        predictions = self.outputs(dense_2)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model


class VGG_16_3D(VGG):
    dims = 3

    def create_model(self):
        self.init_dimensionality(self.dims)
        self.check_minimum_input_size(self.inputs, [24, 24, 24])

        conv_1 = Conv3D(
            filters=64, kernel_size=[3, 3, 3], strides=[1, 1, 1], padding="same", activation="relu",
        )(self.inputs)

        conv_2 = Conv3D(
            filters=64, kernel_size=[3, 3, 3], strides=[1, 1, 1], padding="same", activation="relu",
        )(conv_1)

        vgg_block_1 = self.get_VGG_block(conv_2, 128, 2)
        vgg_block_1 = self.make_dropout_layer(vgg_block_1)
        vgg_block_2 = self.get_VGG_block(vgg_block_1, 256, 3)
        vgg_block_2 = self.make_dropout_layer(vgg_block_2)
        vgg_block_3 = self.get_VGG_block(vgg_block_2, 512, 3)
        vgg_block_3 = self.make_dropout_layer(vgg_block_3)
        vgg_block_4 = self.get_VGG_block(vgg_block_3, 512, 3)

        stride_size = self.get_corrected_stride_size(vgg_block_4, [2, 2, 2], [2, 2, 2])
        pooling_1 = MaxPooling3D(pool_size=[2, 2, 2], strides=stride_size, padding="valid",)(
            vgg_block_4
        )
        flatten_1 = Flatten()(pooling_1)
        flatten_1 = self.make_dropout_layer(flatten_1)
        dense_1 = Dense(4096, activation="relu")(flatten_1)
        dense_2 = Dense(4096, activation="relu")(dense_1)
        predictions = self.outputs(dense_2)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model


class VGG_19_2D(VGG):
    dims = 2

    def create_model(self):
        self.init_dimensionality(self.dims)
        self.check_minimum_input_size(self.inputs, [24, 24])

        conv_1 = Conv2D(
            filters=64, kernel_size=[3, 3], strides=[1, 1], padding="same", activation="relu",
        )(self.inputs)

        conv_2 = Conv2D(
            filters=64, kernel_size=[3, 3], strides=[1, 1], padding="same", activation="relu",
        )(conv_1)

        vgg_block_1 = self.get_VGG_block(conv_2, 128, 2)
        vgg_block_1 = self.make_dropout_layer(vgg_block_1)
        vgg_block_2 = self.get_VGG_block(vgg_block_1, 256, 4)
        vgg_block_2 = self.make_dropout_layer(vgg_block_2)
        vgg_block_3 = self.get_VGG_block(vgg_block_2, 512, 4)
        vgg_block_3 = self.make_dropout_layer(vgg_block_3)
        vgg_block_4 = self.get_VGG_block(vgg_block_3, 512, 4)

        stride_size = self.get_corrected_stride_size(vgg_block_4, [2, 2], [2, 2])
        pooling_1 = MaxPooling2D(pool_size=[2, 2], strides=stride_size, padding="valid",)(
            vgg_block_4
        )
        flatten_1 = Flatten()(pooling_1)
        flatten_1 = self.make_dropout_layer(flatten_1)
        dense_1 = Dense(4096, activation="relu")(flatten_1)
        dense_2 = Dense(4096, activation="relu")(dense_1)
        predictions = self.outputs(dense_2)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model


class VGG_19_3D(VGG):
    dims = 3

    def create_model(self):
        self.init_dimensionality(self.dims)
        self.check_minimum_input_size(self.inputs, [24, 24, 24])

        conv_1 = Conv3D(
            filters=64, kernel_size=[3, 3, 3], strides=[1, 1, 1], padding="same", activation="relu",
        )(self.inputs)

        conv_2 = Conv3D(
            filters=64, kernel_size=[3, 3, 3], strides=[1, 1, 1], padding="same", activation="relu",
        )(conv_1)

        vgg_block_1 = self.get_VGG_block(conv_2, 128, 2)
        vgg_block_1 = self.make_dropout_layer(vgg_block_1)
        vgg_block_2 = self.get_VGG_block(vgg_block_1, 256, 4)
        vgg_block_2 = self.make_dropout_layer(vgg_block_2)
        vgg_block_3 = self.get_VGG_block(vgg_block_2, 512, 4)
        vgg_block_3 = self.make_dropout_layer(vgg_block_3)
        vgg_block_4 = self.get_VGG_block(vgg_block_3, 512, 4)

        stride_size = self.get_corrected_stride_size(vgg_block_4, [2, 2, 2], [2, 2, 2])
        pooling_1 = MaxPooling3D(pool_size=[2, 2, 2], strides=stride_size, padding="valid",)(
            vgg_block_4
        )
        flatten_1 = Flatten()(pooling_1)
        flatten_1 = self.make_dropout_layer(flatten_1)
        dense_1 = Dense(4096, activation="relu")(flatten_1)
        dense_2 = Dense(4096, activation="relu")(dense_1)
        predictions = self.outputs(dense_2)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model



import numpy as np
import pandas as pd
import tensorflow as tf

from PrognosAIs.IO import utils


class LabelLoader:
    def __init__(
        self,
        label_file: str,
        filter_missing: bool = False,
        missing_value: int = -1,
        make_one_hot: bool = False,
        new_root_path: str = None,
    ) -> None:
        """
        Create a label loader, that can load the image paths and labels
        from a text file to be used for a data generator

        Args:
            label_file: The label file from which to read the labels
            filter_missing: Whether missing values should be masked when generating one hot labels and class weights
            missing_value: If filter_missing is True, this value is used to mask
            make_one_hot: Whether labels should be transformed to one hot labels
            new_root_path: If you want to move the files, this will be the new root path
        """

        self.label_file = label_file
        self.label_data = pd.read_csv(label_file, sep="\t", header=0, index_col=0)
        self._original_label_data = self.label_data.copy(deep=True)

        self.filter_missing = filter_missing
        self.missing_value = missing_value
        self.make_one_hot = make_one_hot
        self.one_hot_encoded = False
        self.new_root_path = new_root_path

        self.total_weight_sum = 1.0

        if self.new_root_path is not None:
            self.replace_root_path()

        if self.make_one_hot:
            self.encode_labels_one_hot()

        return

    def get_labels(self) -> list:
        """Get all labels of all samples

        Args:
            None
        Returns:
            labels: List of labels
        """
        labels = np.squeeze(self.label_data.values)

        if isinstance(labels, np.ndarray) and labels.size > 1:
            labels = labels.tolist()
        elif isinstance(labels, np.ndarray):
            # Otherwise if it is 1 element it will remove the list,
            # and return only a string
            labels = [labels.tolist()]
        else:
            labels = [labels]

        return labels

    def get_samples(self) -> list:
        """Get all labels of all samples

        Args:
            None
        Returns:
            samples: List of samples
        """

        return self.label_data.index.to_list()

    def get_data(self) -> dict:
        """Get all data from the label file

        Args:
            None
        Returns:
            data: Dictionary mapping each sample to each label
        """
        return self.label_data.to_dict(orient="index")

    def get_label_from_sample(self, sample: str) -> dict:
        """Get label from a sample

        Args:
            sample: The sample from which to get the label
        Returns:
            label: Label of the sample
        """

        return self.label_data.loc[sample].to_dict()

    def get_label_categories(self) -> list:
        """Get categories of labels

        Args:
            None
        Returns:
            label_categories: Category names
        """

        return self.label_data.columns.to_numpy(copy=True).tolist()

    def get_labels_from_category(self, category_name: str) -> list:
        """Get labels of a specific category/class

        Args:
            category_name: Name of the category/class to get

        Returns:
            list: Labels of the category
        """

        return self.label_data[category_name].to_numpy(copy=True).tolist()

    def get_original_labels_from_category(self, category_name: str) -> list:
        """Get original labels of a specific category/class

        Args:
            category_name: Name of the category/class to get

        Returns:
            list: Original labels of the category
        """

        return self._original_label_data[category_name].to_numpy(copy=True).tolist()

    def get_label_category_type(self, category_name: str) -> type:
        """Get the type of a label of a specific category/class

        Args:
            category_name: Name of the category/class to get type of

        Returns:
            type: Type of the labels of the category
        """

        category_label_type = type(self.label_data[category_name][0])

        return category_label_type

    def get_original_label_category_type(self, category_name: str) -> type:
        """Get the original type of a label of a specific category/class

        Args:
            category_name: Name of the category/class to get type of

        Returns:
            type: Type of the labels of the category
        """

        category_label_type = type(self._original_label_data[category_name][0])

        return category_label_type

    def encode_labels_one_hot(self) -> None:
        """Encode sample labels as one hot

        Args:
            None
        Returns:
            None
        """

        if self.one_hot_encoded:
            return

        label_categories = self.get_label_categories()

        for i_label_category in label_categories:
            category_type = self.get_label_category_type(i_label_category)
            if np.issubdtype(category_type, np.integer):
                category_labels = self.get_labels_from_category(i_label_category)
                category_labels = np.asarray(category_labels)
                labels_min = np.amin(category_labels)
                if self.filter_missing and labels_min == self.missing_value:
                    labels_min = np.amin(category_labels[category_labels != self.missing_value])
                N_labels = np.amax(category_labels) - labels_min + 1
                category_labels -= labels_min
                one_hot_labels = tf.one_hot(category_labels, N_labels, dtype=tf.int8).numpy()
                # We need to replace the one hot labels with the value again
                missing_label_index = np.sum(one_hot_labels, axis=-1)
                one_hot_labels[missing_label_index == 0] = self.missing_value
                self.label_data[i_label_category] = one_hot_labels.tolist()

        self.one_hot_encoded = True

        return

    def replace_root_path(self) -> None:
        """Replace the root path of the sample files in case they have been
        moved to a different a different directory.

        Args:
            new_root_path: Path in which the files are now located
        Returns:
            None
        """

        if self.new_root_path is not None:
            samples = self.get_samples()

            new_root_path = utils.normalize_path(self.new_root_path)

            for i_i_sample, i_sample in enumerate(samples):
                old_root_path = utils.get_file_path(i_sample)
                samples[i_i_sample] = i_sample.replace(old_root_path, new_root_path)

            self.label_data.index = samples

            label_categories = self.get_label_categories()
            for i_label_category in label_categories:
                if self.get_label_category_type(i_label_category) is str:
                    category_labels = self.get_labels_from_category(i_label_category)

                    for i_i_label, i_label in enumerate(category_labels):
                        old_root_path = utils.get_file_path(i_label)
                        category_labels[i_i_label] = i_label.replace(old_root_path, new_root_path)

                    self.label_data[i_label_category] = category_labels
            return

    def get_class_weights(self, json_serializable=False) -> dict:
        """ Get class weights for unbalanced labels

        Args:
            None
        Returns:
            Scaled_weights: the weights for each class of each label category, scaled
                such that the total weights*number of samples of each class
                approximates the total number of samples
        """

        out_scaled_weights = {}

        for i_label_category in self.get_label_categories():
            category_type = self.get_original_label_category_type(i_label_category)
            if np.issubdtype(category_type, np.integer):
                category_labels = self.get_original_labels_from_category(i_label_category)
                category_labels = np.asarray(category_labels)
                if self.filter_missing:
                    category_labels = category_labels[category_labels != self.missing_value]
                if self.make_one_hot:
                    category_labels = category_labels - np.amin(category_labels)
                classes, counts = np.unique(category_labels, return_counts=True)
                N_samples = len(category_labels)

                weights = N_samples / (counts * len(classes))
                if json_serializable:
                    classes = [str(i_class) for i_class in classes]
                # Dictionary for each category, with dictionary of weight for each class in that category
                # This allows for easy input with tensorflow
                out_scaled_weights[i_label_category] = dict(zip(classes, weights))
            else:
                out_scaled_weights[i_label_category] = None

        return out_scaled_weights

    def get_number_of_classes_from_category(self, category_name: str) -> int:
        """ Get number of classes for a label category

        Args:
            category_name: Category to get number of classes for
        Returns:
            number_of_classes: The number of classes for the category
        """

        category_type = self.get_original_label_category_type(category_name)
        if np.issubdtype(category_type, np.integer):
            category_labels = self.get_original_labels_from_category(category_name)
            category_labels = np.asarray(category_labels)
            if self.filter_missing:
                category_labels = category_labels[category_labels != self.missing_value]
            classes = np.unique(category_labels)
            number_of_classes = len(classes)
        else:
            number_of_classes = -1

        return number_of_classes

    def get_number_of_classes(self) -> dict:
        """ Get number of classes for all categories

        Args:
            None
        Returns:
            number_of_classes: The number of classes for each category
        """

        number_of_classes = {}

        for i_label_category in self.get_label_categories():
            number_of_classes[i_label_category] = self.get_number_of_classes_from_category(
                i_label_category
            )

        return number_of_classes

    def get_number_of_samples(self) -> int:
        """ Get number of samples

        Args:
            None
        Returns:
            number_of_samples: The number of samples
        """

        return len(self.get_samples())



import importlib
import os
import shutil
import logging
import sys

from pathlib import Path

import numba.cuda
import psutil
import tensorflow as tf

from slurmpie import slurmpie


def create_directory(file_path, exist_ok=True):
    if not os.path.exists(file_path):
        os.makedirs(file_path, exist_ok=exist_ok)


def delete_directory(file_path):
    if os.path.exists(file_path):
        shutil.rmtree(file_path)


def copy_directory(original_directory, out_directory):
    shutil.copytree(original_directory, out_directory, dirs_exist_ok=True)


def get_root_name(file_path):
    return os.path.basename(os.path.normpath(file_path))


def get_file_name_from_full_path(file_path):
    return os.path.basename(os.path.normpath(file_path))


def get_file_name(file_path, file_extension):
    root_name = get_root_name(file_path)
    if file_extension[0] != ".":
        file_extension = ".".join(file_extension)
    return root_name.split(file_extension)[0]


def find_files_with_extension(file_path, file_extension):
    if file_extension[0] == ".":
        file_extension = file_extension[1:]
    return sorted(
        f_path.path
        for f_path in os.scandir(file_path)
        if (f_path.is_file() and file_extension in "".join(Path(f_path.name).suffixes))
    )


def get_parent_directory(file_path):
    return os.path.dirname(os.path.normpath(os.path.abspath(file_path)))


def get_file_path(file_path):
    file_name = get_root_name(file_path)
    file_path = file_path.split(file_name)[0]
    return normalize_path(file_path)


def normalize_path(path):
    if path[-1] == os.sep:
        path = path[:-1]
    return path


def get_number_of_cpus():
    return len(os.sched_getaffinity(0))


def get_subdirectories(root_dir: str) -> list:
    return [f_path.path for f_path in os.scandir(root_dir) if f_path.is_dir()]


def get_available_ram(used_memory: int = 0) -> int:
    """
    Get the available RAM in bytes.

    Returns:
        int: available in RAM in bytes
    """
    slurm_mem = slurmpie.System().get_job_memory()
    if slurm_mem is None:
        available_ram = psutil.virtual_memory().available
    else:
        # Convert from megabytes to bytes (*1024*1024)
        slurm_mem *= 1048576
        available_ram = slurm_mem - used_memory
    return available_ram


def get_dir_size(root_dir):
    """Returns total size of all files in dir (and subdirs)"""
    root_directory = Path(os.path.normpath(root_dir))
    return sum(f.stat().st_size for f in root_directory.glob("**/*") if f.is_file())


def get_gpu_compute_capability(gpu: tf.config.PhysicalDevice) -> tuple:
    try:
        gpu_number = int(gpu.name.split(":")[-1])
        cuda_device = numba.cuda.select_device(gpu_number)
        cuda_capability = cuda_device.compute_capability
        cuda_device.reset()
    except numba.cuda.cudadrv.error.CudaSupportError:
        # We do not actually have a cuda device
        cuda_capability = (0, 0)
    return cuda_capability


def gpu_supports_float16(gpu: tf.config.PhysicalDevice) -> bool:
    gpu_compute_capability = get_gpu_compute_capability(gpu)
    # Float16 support is supported with at least compute capability 5.3
    supports_float16 = (gpu_compute_capability[0] == 5 and gpu_compute_capability[1] >= 3) or (
        gpu_compute_capability[0] > 5
    )
    return supports_float16


def gpu_supports_mixed_precision(gpu: tf.config.PhysicalDevice) -> bool:
    gpu_compute_capability = get_gpu_compute_capability(gpu)
    # Mixed precision has benefits on compute 7.5 and higher
    return gpu_compute_capability[0] >= 7 and gpu_compute_capability[1] >= 5


def get_gpu_devices() -> list:
    return tf.config.list_physical_devices("GPU")


def get_number_of_gpu_devices() -> int:
    return len(tf.config.list_physical_devices("GPU"))


def get_cpu_devices() -> list:
    return tf.config.list_physical_devices("CPU")


def get_number_of_slurm_nodes() -> int:
    if "SLURM_JOB_NUM_NODES" in os.environ:
        number_of_slurm_nodes = int(os.environ["SLURM_JOB_NUM_NODES"])
    else:
        number_of_slurm_nodes = 0

    return number_of_slurm_nodes


def load_module_from_file(module_path):
    if module_path is None:
        return None
    class_name = get_root_name(module_path).split(".")[0]
    module_file_spec = importlib.util.spec_from_file_location(class_name, module_path,)
    module = importlib.util.module_from_spec(module_file_spec)
    module_file_spec.loader.exec_module(module)

    return module


def setup_logger():
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s prognosais %(levelname)-1s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        stream=sys.stdout,
    )



HDF5_EXTENSION = "hdf5"
HDF_SEPARATOR = "/"
MODEL_SUBFOLDER = "MODEL"
CLASS_WEIGHT_FILE = "class_weights.json"
TRAIN_DS_NAME = "train"
VALIDATION_DS_NAME = "validation"
TEST_DS_NAME = "test"

# Used to determine whether a dataset can safely fit into
# The available memory
MEM_SAFETY_FACTOR = 1.05

# this constant is used everywhere to check whether
# we should automagically apply that thing
AUTO = "AUTO"


FEATURE_INDEX = "feature"
LABEL_INDEX = "label"



import tensorflow as tf

from tensorflow.keras.losses import CategoricalCrossentropy
from tensorflow.keras.losses import Loss
from tensorflow.python.framework import ops
from tensorflow.python.keras.utils import losses_utils
from tensorflow.python.ops import math_ops


class MaskedCategoricalCrossentropy(CategoricalCrossentropy):
    """Caterogical crossentropy loss which takes into account missing values."""

    def __init__(
        self,
        name: str = "masked_categorical_crossentropy",
        class_weight: dict = None,
        mask_value: int = -1,
        **kwargs,
    ) -> None:
        r"""
        Caterogical crossentropy loss which takes into account missing values.

        For the samples with masked values a cross entropy of 0 will be used, for the other samples
        the standard cross entropy loss will be calculated

        Args:
            name (str): Optional name for the op
            class_weight (dict): Weights for each class
            mask_value (int): The value that indicates that a sample is missing
            **kwargs:  arguments to pass the default CategoricalCrossentropy loss
        """
        super().__init__(name=name, **kwargs)
        self.dtype = tf.dtypes.as_dtype(tf.keras.backend.floatx())

        self.mask_value = mask_value
        self.original_class_weight = class_weight
        if class_weight is not None:
            # Got a dict with class weights for different classes
            # Need to prepare this for tensorflow computation
            self.class_weight = tf.constant(
                [
                    float(v)
                    for k, v in sorted(class_weight.items(), key=lambda item: float(item[0]))
                ],
            )
            print(self.class_weight)
        else:
            self.class_weight = None

    def is_unmasked_sample(self, y_true: tf.Tensor) -> tf.Tensor:
        """
        Get whether the samples are unmasked (i.e. have real label data).

        Args:
            y_true (tf.Tensor): Tensor of the true labels

        Returns:
            tf.Tensor: Tensor of 0s and 1s indicating whether that sample is unmasked.
        """
        unmasked_samples = tf.reduce_all(tf.math.not_equal(y_true, self.mask_value), axis=-1)
        return math_ops.cast(unmasked_samples, self.dtype)

    def call(self, y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
        """
        Obtain the masked categorical crossentropy loss for each sample.

        Args:
            y_true (tf.Tensor): Ground-truth labels, one-hot encoded
                (batch_size, N_1, N_2, .... N_d) tensor, with N_d the number of outputs
            y_pred (tf.Tensor): Predictions one-hot encoded, for example from softmax,
                (batch_size, N_1, N_2, .... N_d) tensor, with N_d the number of outputs

        Returns:
            tf.Tensor: The masked categorial crossentropy loss for each sample, has rank
                one less than the inputs tensors
        """
        y_pred = ops.convert_to_tensor(y_pred)
        y_pred = tf.cast(y_pred, self.dtype)
        y_true = math_ops.cast(y_true, y_pred.dtype)
        loss = super().call(y_true, y_pred) * self.is_unmasked_sample(y_true)
        if self.class_weight is not None:
            # Correction factor is need to keep loss total the same as
            # for case without class weights
            y_true_indexes = tf.math.argmax(y_true, axis=-1)
            weights = math_ops.cast(tf.gather(self.class_weight, y_true_indexes), loss.dtype)
            loss *= weights
        return loss

    def __call__(
        self, y_true: tf.Tensor, y_pred: tf.Tensor, sample_weight: tf.Tensor = None,
    ) -> tf.Tensor:
        """
        Obtain the total masked categorical crossentropy loss for the batch.

        Args:
            y_true (tf.Tensor): Ground-truth labels, one-hot encoded
                (batch_size, N_1, N_2, .... N_d) tensor, with N_d the number of outputs
            y_pred (tf.Tensor): Predictions one-hot encoded, for example from softmax,
                (batch_size, N_1, N_2, .... N_d) tensor, with N_d the number of outputs
            sample_weight (tf.Tensor): Sample weight for each indidvidual label
                to be used in reduction of sample loss to overal batch loss

        Returns:
            tf.Tensor: The total masked categorial crossentropy loss, scalar tensor with rank 0
        """
        losses = self.call(y_true, y_pred)
        unmasked_samples = self.is_unmasked_sample(y_true)
        weight_correction = unmasked_samples * tf.math.divide_no_nan(
            tf.cast(tf.gather(tf.shape(y_true), 0), self.dtype),
            tf.math.count_nonzero(unmasked_samples, dtype=self.dtype),
        )

        if sample_weight is not None:
            sample_weight *= weight_correction
        else:
            sample_weight = weight_correction

        return losses_utils.compute_weighted_loss(
            losses, sample_weight, reduction=self._get_reduction(),
        )

    def get_config(self) -> dict:
        """
        Get the configuration of the loss.

        Returns:
            dict: Configuration parameters of the loss
        """
        config = super().get_config()
        config["mask_value"] = self.mask_value
        if self.original_class_weight is not None:
            to_write_class_weight = {}
            for key in self.original_class_weight.keys():
                to_write_class_weight[str(key)] = str(self.original_class_weight[key])
            config["class_weight"] = to_write_class_weight

        return config


class DICE_loss(Loss):
    """Loss class for the Sørensen–Dice coefficient."""

    def __init__(
        self,
        name: str = "dice_loss",
        weighted: bool = False,
        foreground_only: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(name=name, **kwargs)

        self.weighted = weighted
        self.foreground_only = foreground_only

    # TODO: Add masked DICE loss as well
    def call(self, y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
        r"""
        Calculate the DICE loss.

        This functions calculates the DICE loss defined as:

        .. math::
            1 - 2 * \frac{|A \cap B|}{|A| + |B|}

        When no positive labels are found in both A and B the loss
        returns 0 by default.
        The loss works both for one-hot predicted labels and binary labels.

        Args:
            y_true (tf.Tensor): The ground truth labels, shape: (batch_size, N_1, N_2 ... N_d)
                where N_d is the number of channels (can be 1). For a 3D tensor with 1 channel (binary class)
                and batch size of 1 it will have a shape of (1, N_1, N_2, N_3, 1)
            y_pred (tf.Tensor): The predicted labels. shape: (batch_size, N_1, N_2 ... N_d)
                where N_d is the number of channels. When a binary prediction is done (last activation function
                is sigmoid), N_d = 1. When one-hot prediction are done (last activation function is softmax)
                N_d = number of classes

        Returns:
            tf.Tensor: Tensor of length batch_size with the DICE loss for each sample
        """
        y_pred = ops.convert_to_tensor(y_pred)
        # TODO FIX THIS CLIPPING< TO AVOID NAN IN FLOAT16
        y_pred = tf.clip_by_value(y_pred, 0, 1)
        y_true = math_ops.cast(y_true, y_pred.dtype)

        to_reduce_axis = tf.range(1, tf.rank(y_true) - 1, 1)
        numerator = tf.math.reduce_sum(y_true * y_pred, axis=to_reduce_axis)
        denominator = tf.math.reduce_sum(y_true + y_pred, axis=to_reduce_axis)
        # We take 1 minus the dice to get the loss instead of the score
        # We multiply by product of y_true to get loss of 0 when class is not present
        dice_loss_per_class = (
            1.0 - 2.0 * tf.math.divide_no_nan(numerator, denominator)
        ) * tf.math.reduce_max(y_true, axis=to_reduce_axis)

        if self.weighted:
            samples_per_class = tf.math.reduce_sum(y_true, axis=to_reduce_axis)
            total_samples = tf.math.reduce_sum(samples_per_class, axis=-1, keepdims=True)
            # total_samples = tf.broadcast_to(total_samples, tf.shape(samples_per_class))
            weight_per_class = tf.math.divide_no_nan(total_samples, samples_per_class)
            # Apply normalization factor
            weight_per_class = tf.math.divide(
                weight_per_class, tf.math.reduce_sum(weight_per_class, axis=-1, keepdims=True)
            )

            dice_loss_per_class = dice_loss_per_class * weight_per_class

        if self.foreground_only:
            dice_loss_per_class = dice_loss_per_class[..., 1:]

        # Finally we get mean over classes
        return tf.math.reduce_mean(dice_loss_per_class, axis=-1)

    def get_config(self) -> dict:
        """
        Get the configuration of the loss.

        Returns:
            dict: configuration of the loss
        """
        config = super().get_config()
        config["weighted"] = self.weighted
        config["foreground_only"] = self.foreground_only
        return config


class CoxLoss(Loss):
    """Cox loss as defined in https://arxiv.org/pdf/1606.00931.pdf."""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.dtype = tf.dtypes.as_dtype(tf.keras.backend.floatx())

    def call(self, y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
        r"""
        Calculate the cox loss.

        Args:
            y_true (tf.Tensor): Tensor of shape (batch_size, 2), with the first index
                containing whether and event occurred for each sample, and the second
                index containing the time to event, or follow-up time if no event
                has occurred
            y_pred (tf.Tensor): The :math:`\hat{h}_{\sigma\}` as predicted by the network

        Returns:
            tf.Tensor: The cox loss for each sample in the batch
        """
        # We clip the predictions to make sure we dont get overflow problems
        y_pred = tf.clip_by_value(ops.convert_to_tensor(y_pred), 0, tf.math.log(self.dtype.max) - 1)
        y_true = math_ops.cast(y_true, y_pred.dtype)
        events_occurred = y_true[:, 0]
        time_to_events = y_true[:, 1]
        epsilon = 1e-6

        # This creates a matrix comparing each time to event with all others
        risk_set = tf.cast(
            tf.math.greater_equal(time_to_events[None, :], time_to_events[:, None]), self.dtype,
        )

        # Get the exponential hazard for all sample_locations
        # We clip in case of very large predicted values to ensure that
        # We will never get an overflow
        exp_hazard = tf.math.exp(y_pred)

        # Get risk only for samples for which event did not occur yet
        # Calculate the sum over the risk sets
        exp_hazard_risk_set = tf.squeeze(tf.math.reduce_sum(exp_hazard[None, :] * risk_set, axis=1))

        # Get log of sum over risk set, with epsilon in case it is 0
        # Get difference between prediction and log of risk set
        hazard_difference = tf.squeeze(y_pred) - tf.math.log(exp_hazard_risk_set + epsilon)

        # Only values for patients with events
        # We want to minimize a function, so need to make loss positive
        return tf.math.negative(hazard_difference * events_occurred)

    def get_config(self) -> dict:
        """
        Get the configuration of the loss.

        Returns:
            dict: configuration of the loss
        """
        return super().get_config()



import argparse
import os
import sys

from types import ModuleType
from typing import Dict
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd
import SimpleITK as sitk
import tensorflow as tf

from tensorflow.keras.models import load_model

import PrognosAIs.Constants
import PrognosAIs.IO.utils as IO_utils
import PrognosAIs.Model.Losses
import PrognosAIs.Model.Parsers as ModelParsers

from PrognosAIs.IO import ConfigLoader
from PrognosAIs.IO import DataGenerator


class Evaluator:
    def __init__(self, model_file, data_folder, config_file, output_folder) -> None:
        self.EVALUATION_FOLDER = "Results"
        self.data_folder = data_folder
        self.model_file = model_file
        self.predictions = None
        self.metrics = {}
        self.metric_functions = None
        self.sample_predictions = None
        self.sample_labels = None
        self.sample_names = None
        self.sample_metrics = None
        self.output_folder = os.path.join(output_folder, self.EVALUATION_FOLDER)
        IO_utils.create_directory(self.output_folder)

        self.config = ConfigLoader.ConfigLoader(config_file)
        self.batch_size = self.config.get_batch_size()

        self.custom_module = IO_utils.load_module_from_file(
            self.config.get_custom_definitions_file(),
        )
        self.model = self.load_model(self.model_file, self.custom_module)

        self.init_model_parameters()

        self.data_generator = self.init_data_generators()
        self.dataset_names = list(self.data_generator.keys())
        self.sample_metadata = self.data_generator[self.dataset_names[0]].get_feature_metadata()
        self.label_data_generator = self.init_label_generators()

        self.metric_functions = self.get_to_evaluate_metrics()

        self.image_output_labels = self.get_image_output_labels()

    @staticmethod
    def _load_model(model_file: str, custom_objects: dict) -> Tuple[tf.keras.Model, ValueError]:
        """
        Try to load a model, if it doesnt work parse the error.

        Args:
            model_file (str): Location of the model file
            custom_objects (dict): Potential custom objects to use during model loading

        Returns:
            Tuple[tf.keras.Model, ValueError]: The model if successfully loaded, otherwise the error
        """
        try:
            model = load_model(model_file, compile=True, custom_objects=custom_objects)
            out_error = None
        except ValueError as the_error:
            out_error = the_error
            model = None

        return model, out_error

    @staticmethod
    def _fake_fit(model: tf.keras.Model) -> tf.keras.Model:
        """
        Fit of the model on fake date to properly initialize the model.

        Args:
            model (tf.keras.Model): Model to initialize

        Returns:
            tf.keras.Model: Initalized model.
        """
        input_shapes = model.input_shape
        output_shapes = model.output_shape

        if isinstance(input_shapes, dict):
            fake_input = {}
            for i_key, i_value in input_shapes.items():
                fake_input[i_key] = tf.random.uniform((1,) + i_value[1:])
        else:
            fake_input = tf.random.uniform((1,) + input_shapes[1:])

        if isinstance(output_shapes, list):
            fake_output = []
            for i_value in output_shapes:
                fake_output.append(tf.random.uniform((1,) + i_value[1:]))
        else:
            fake_output = tf.random.uniform((1,) + output_shapes[1:])

        # We are not currently doing it as number of epochs is not very important in evaluation
        # But you could get the run number of epochs from the loaded model via:
        # model.optimizer.iterations and divided this by the number of steps
        # in each epoch (which is #samples/batch size)

        model.fit(fake_input, fake_output, verbose=0)
        return model

    @staticmethod
    def load_model(model_file: str, custom_module: ModuleType = None) -> tf.keras.Model:
        """
        Load the model, including potential custom losses.

        Args:
            model_file (str): Location of the model file
            custom_module (ModuleType): Custom module from which to load losses or metrics

        Raises:
            error: If the model could not be loaded
                and the problem is not due to a missing loss or metric function.

        Returns:
            tf.keras.Model: The loaded model
        """
        # type hint for mypy
        custom_objects: Dict[str, Union[tf.keras.losses.Loss.tf.keras.metrics.Metric]] = {}
        model_is_loaded = False
        model = None
        while not model_is_loaded:
            model, error = Evaluator._load_model(model_file, custom_objects)
            if model is not None:
                model_is_loaded = True
                break

            err_msg = error.args[0]
            custom_object_name = err_msg.split(":")[-1].strip()

            if "Unknown loss function" in str(err_msg):
                custom_object = getattr(PrognosAIs.Model.Losses, custom_object_name, None)
                if custom_object is None:
                    custom_object = getattr(custom_module, custom_object_name)

            elif "Unknown metric function" in str(err_msg):
                custom_object = getattr(PrognosAIs.Model.Metrics, custom_object_name, None)
                if custom_object is None:
                    custom_object = getattr(custom_module, custom_object_name)
            elif "Unknown optimizer" in str(err_msg):
                custom_object = getattr(custom_module, custom_object_name)
            else:
                raise error
            custom_objects.update({custom_object_name: custom_object})

        # There is an "update" in TF2.2 which does not make metrics available by default
        # from loaded models, see https://github.com/tensorflow/tensorflow/issues/37990
        # Therefore we do a fake "fit" round to make everything available
        return Evaluator._fake_fit(model)

    def _init_data_generators(self, labels_only: bool) -> dict:
        """
        Initialize data generators for all sample folders.

        Args:
            labels_only (bool): Whether to only load labels

        Returns:
            dict: initalized data generators
        """
        sub_folders = IO_utils.get_subdirectories(self.data_folder)
        data_generators = {}
        for i_sub_folder in sub_folders:
            folder_name = IO_utils.get_root_name(i_sub_folder)
            if (
                folder_name == PrognosAIs.Constants.TRAIN_DS_NAME
                and not self.config.get_evaluate_train_set()
            ):
                continue
            data_generators[folder_name] = DataGenerator.HDF5Generator(
                i_sub_folder,
                self.batch_size,
                shuffle=self.config.get_shuffle_evaluation(),
                drop_batch_remainder=False,
                labels_only=labels_only,
            )
        return data_generators

    def init_data_generators(self) -> dict:
        """
        Initialize the data generators.

        Returns:
            dict: DataGenerator for each subfolder of samples
        """

        return self._init_data_generators(False)

    def init_label_generators(self) -> dict:
        """
        Initialize the data generators which only give labels.

        Returns:
            dict: DataGenerator for each subfolder of samples
        """
        return self._init_data_generators(False)

    def init_model_parameters(self) -> None:
        """
        Initialize the parameters from the model.
        """
        self.output_names = self.model.output_names
        self.number_of_outputs = len(self.output_names)
        if self.number_of_outputs == 1:
            self.output_shapes = [self.model.output_shape]
        else:
            self.output_shapes = self.model.output_shape

        self.output_classes = {}
        self.one_hot_outputs = {}
        for i_output_index, i_output_name in enumerate(self.output_names):
            self.output_classes[i_output_name] = self.output_shapes[i_output_index][-1]

            if self.output_shapes[i_output_index][-1] > 1:
                self.one_hot_outputs[i_output_name] = True
            else:
                self.one_hot_outputs[i_output_name] = False

        self.input_names = self.model.input_names
        self.number_of_inputs = len(self.input_names)

        model_input_shape = self.model.input_shape
        if isinstance(model_input_shape, dict):
            self.input_shapes = list(model_input_shape.values())
        elif self.number_of_inputs == 1:
            self.input_shapes = [model_input_shape]
        else:
            self.input_shapes = model_input_shape.values()

    def get_image_output_labels(self) -> dict:
        """
        Whether an output label is a simple class, the label is actually an image.

        Returns:
            dict: Output labels that are image outputs
        """
        image_outputs_labels = {}

        for i_output_name, i_output_shape in zip(self.output_names, self.output_shapes):
            for i_input_name, i_input_shape in zip(self.input_names, self.input_shapes):
                # It is an image of a certain input of the output has as many dimension
                # and the size of each dimension is equal to the input size
                # minus the batch dimension and number of classes

                equal_dimensions = len(i_input_shape) == len(i_output_shape)
                equal_size = i_input_shape[1:-1] == i_output_shape[1:-1]
                if equal_dimensions and equal_size:
                    image_outputs_labels[i_output_name] = i_input_name

        return image_outputs_labels

    def get_real_labels(self) -> dict:
        real_labels = {}
        for i_generator_name in self.data_generator.keys():
            real_labels[i_generator_name] = self.get_real_labels_of_sample_subset(i_generator_name)

        return real_labels

    def get_real_labels_of_sample_subset(self, subset_name: str) -> dict:
        """
        Get the real labels corresponding of all samples from a subset.

        Args:
            subset_name (str): Name of subset to get labels for

        Returns:
            dict: Real labels for each dataset and output
        """
        real_labels = {}
        subset_generator = self.data_generator[subset_name].get_tf_dataset()
        for i_output_name in self.output_names:
            real_labels[i_output_name] = np.concatenate(
                [i_label_batch[1][i_output_name] for i_label_batch in subset_generator], axis=0,
            )
        return real_labels

    def _combine_config_and_model_metrics(self, model_metrics: dict, config_metrics: dict) -> dict:
        """
        Combine the metrics specified in the model and those specified in the config.

        Args:
            model_metrics (dict): Metrics as defined by the model
            config_metrics (dict): Metrics defined in the config

        Returns:
            dict: Combined metrics
        """
        metric_functions = {}
        for i_output_name in self.output_names:
            if i_output_name in model_metrics and i_output_name in config_metrics:
                metric_functions[i_output_name] = (
                    model_metrics[i_output_name] + config_metrics[i_output_name]
                )
            elif i_output_name in model_metrics:
                metric_functions[i_output_name] = model_metrics[i_output_name]
            elif i_output_name in config_metrics:
                metric_functions[i_output_name] = config_metrics[i_output_name]

        return metric_functions

    def get_to_evaluate_metrics(self) -> dict:
        """
        Get the metrics functions which should be evaluated.

        Returns:
            dict: Metric function to be evaluated for the different outputs
        """
        # We get both the metrics that are specified in the config
        # As well as the metrics from the model
        metric_parser = ModelParsers.MetricParser(
            self.config.get_evaluation_metric_settings(),
            self.output_names,
            self.config.get_custom_definitions_file(),
        )
        model_metrics = self.model.metrics

        # First N metrics are always losses, we dont want to evaluate those
        if self.number_of_outputs > 1 and len(model_metrics) > self.number_of_outputs:
            # When we have multiple outputs we need to do + 1 because we also
            # have the total loss
            model_metrics = model_metrics[self.number_of_outputs + 1 :]
        elif len(model_metrics) > self.number_of_outputs:
            model_metrics = model_metrics[self.number_of_outputs :]
        else:
            # In this case there are no metrics, only losses
            model_metrics = []

        model_metrics = metric_parser.convert_metrics_list_to_dict(model_metrics)
        config_metrics = metric_parser.get_metrics()
        config_metrics = metric_parser.convert_metrics_list_to_dict(config_metrics)

        return self._combine_config_and_model_metrics(model_metrics, config_metrics)

    def _format_predictions(self, predictions: Union[list, np.ndarray]) -> dict:
        """
        Format the predictions to match them with the output names

        Args:
            predictions (Union[list, np.ndarray]): The predictions from the model

        Raises:
            ValueError: If the predictions do not match with the expected output names

        Returns:
            dict: Output predictions matched with the output names
        """
        if isinstance(predictions, np.ndarray):
            # There is only one output in this case
            predictions = [predictions]

        if len(predictions) != len(self.output_names):
            raise ValueError("The predictions do not match with the output names!")

        out_predictions = {}
        for i_output_name, i_prediction in zip(self.output_names, predictions):
            out_predictions[i_output_name] = i_prediction

        return out_predictions

    def predict(self) -> dict:
        """
        Get predictions from the model

        Returns:
            dict: Predictions for the different outputs of the model for all samples
        """
        if self.predictions is None:
            # We have not yet determined the predictions, first run
            self.predictions = {}
            predictions = {}
            for i_generator_name, i_generator in self.data_generator.items():
                # We go over all generators
                self.predictions[i_generator_name] = {}
                dataset = i_generator.get_tf_dataset()
                final_predictions = {}
                for i_output_name in self.output_names:
                    final_predictions[i_output_name] = []

                for i_batch in dataset:
                    # We have to go over the different predictions step by step
                    # Otherwise will lead to memory leak
                    # The first index in the batch is the sample (the second is the label)
                    batch_prediction = self.model.predict_on_batch(i_batch[0])

                    # Convert to list if we only have one output
                    if isinstance(batch_prediction, np.ndarray):
                        batch_prediction = [batch_prediction]
                    for i_output_name, i_prediction in zip(self.output_names, batch_prediction):
                        final_predictions[i_output_name].append(i_prediction)

                # We create one single list for all predictions that we got
                for i_output_name in self.output_names:
                    final_predictions[i_output_name] = np.concatenate(
                        final_predictions[i_output_name], axis=0
                    )

                predictions[i_generator_name] = final_predictions
            self.predictions = predictions
        return self.predictions

    def evaluate_metrics_from_predictions(self, predictions: dict, real_labels: dict) -> dict:
        """
        Evaluate the metrics based on the model predictions

        Args:
            predictions (dict): Predictions obtained from the model
            real_labels (dict): The true labels of the samples for the different outputs

        Returns:
            dict: The different evaluated metrics
        """

        out_metrics = {}
        for i_output_name, i_output_prediction in predictions.items():
            out_metrics[i_output_name] = {}
            to_evaluate_functions = self.metric_functions[i_output_name]
            for i_function in to_evaluate_functions:
                for i_real_label, i_prediction in zip(
                    real_labels[i_output_name], i_output_prediction
                ):
                    i_real_label = np.asarray(i_real_label)
                    i_prediction = np.asarray(i_prediction)
                    if len(i_real_label.shape) == 1:
                        i_real_label = np.expand_dims(i_real_label, 0)
                        i_prediction = np.expand_dims(i_prediction, 0)

                    i_function.update_state(i_real_label, i_prediction)
                out_metrics[i_output_name][i_function.name] = i_function.result().numpy()
                i_function.reset_states()

        return out_metrics

    def evaluate_metrics(self) -> dict:
        """
        Evaluate all metrics for all samples

        Returns:
            dict: The evaluated metrics
        """
        if self.metrics == {}:
            predictions = self.predict()

            for i_dataset_name, i_predictions in predictions.items():
                real_labels = self.get_real_labels_of_sample_subset(i_dataset_name)
                self.metrics[i_dataset_name] = self.evaluate_metrics_from_predictions(
                    i_predictions, real_labels,
                )
        return self.metrics

    def evaluate_sample_metrics(self) -> dict:
        """
        Evaluate the metrics based on a full sample instead of based on individual batches

        Returns:
            dict: The evaluated metrics
        """
        if self.sample_metrics is None:
            self.sample_metrics = {}
            _, sample_predictions = self.get_sample_predictions_from_patch_predictions()
            _, sample_labels = self.get_sample_labels_from_patch_labels()

            for i_dataset_name, i_predictions in sample_predictions.items():
                self.sample_metrics[i_dataset_name] = self.evaluate_metrics_from_predictions(
                    i_predictions, sample_labels[i_dataset_name],
                )
        return self.sample_metrics

    @staticmethod
    def combine_predictions(
        predictions: np.ndarray, are_one_hot: bool, label_combination_type: str
    ) -> np.ndarray:
        if label_combination_type == "vote":
            if are_one_hot and predictions.ndim >= 2:
                # In case of one-hot we have to sum over the sample dimension
                # We then can just take the argmax to get the real label
                prediction_occurrence = np.sum(np.round(predictions), axis=0)
                majority_vote_index = np.argmax(prediction_occurrence)

                combined_prediction = np.zeros_like(prediction_occurrence)
                combined_prediction[majority_vote_index] = 1
            elif are_one_hot:
                combined_prediction = predictions
            else:
                combined_prediction = np.argmax(np.bincount(predictions))
        elif label_combination_type == "average":
            if are_one_hot and predictions.ndim >= 2:
                combined_prediction = np.mean(predictions, axis=0)
            elif are_one_hot:
                combined_prediction = predictions
            else:
                err_msg = (
                    "Predictions can only be combined when given as probability score"
                    "thus the labels must be one-hot encoded."
                )
                raise ValueError(err_msg)
        else:
            raise ValueError("Unknown combination type")

        return combined_prediction

    def patches_to_sample_image(
        self,
        datagenerator: PrognosAIs.IO.DataGenerator.HDF5Generator,
        filenames: list,
        output_name: str,
        predictions: np.ndarray,
        labels_are_one_hot: bool,
        label_combination_type: str,
    ) -> np.ndarray:

        if not labels_are_one_hot and label_combination_type == "average":
            err_msg = (
                "Predictions can only be combined when given as probability score"
                "thus the labels must be one-hot encoded."
            )
            raise ValueError(err_msg)

        input_name = self.image_output_labels[output_name]
        image_size = self.sample_metadata[input_name]["original_size"]
        transpose_dims = np.arange(len(image_size) - 1, -1, -1)
        number_of_classes = self.output_classes[output_name]
        image_size = np.append(image_size, number_of_classes)

        original_image = np.zeros(image_size)
        number_of_hits = np.zeros(image_size)

        if isinstance(filenames, str):
            filenames = [filenames]

        # TODO REMOVE ONLY FOR IF NOT REALLY PATCHES
        if (
            len(predictions.shape) == len(original_image.shape)
            and predictions.shape[-1] == original_image.shape[-1]
        ):
            predictions = np.expand_dims(predictions, axis=0)

        for i_filename, i_prediction in zip(filenames, predictions):
            i_sample_metadata = datagenerator.get_feature_metadata_from_sample(i_filename)
            i_sample_metadata = i_sample_metadata[input_name]

            in_sample_index_start = np.copy(i_sample_metadata["index"])
            in_sample_index_end = in_sample_index_start + i_sample_metadata["size"]

            # Parts of the patch can be outside of the original image, because of padding
            # Thus here we take only the parts of the patch that are within the original image
            in_sample_index_start[in_sample_index_start < 0] = 0
            sample_indices = tuple(
                slice(*i) for i in zip(in_sample_index_start, in_sample_index_end)
            )

            patch_index_start = np.copy(i_sample_metadata["index"])
            patch_index_end = i_sample_metadata["size"]
            # We also need to cut out the part of the patch that is normally outside of the image
            # we do this here
            patch_index_start[patch_index_start > 0] = 0
            patch_index_start = -1 * patch_index_start
            patch_slices = tuple(slice(*i) for i in zip(patch_index_start, patch_index_end))

            if not labels_are_one_hot:
                i_prediction = i_prediction.astype(np.int32)
                i_prediction = np.eye(number_of_classes)[i_prediction]
            elif label_combination_type == "vote":
                i_prediction = np.round(i_prediction)

            original_image[sample_indices] += i_prediction[patch_slices]
            number_of_hits[sample_indices] += 1

        number_of_hits[number_of_hits == 0] = 1
        if label_combination_type == "vote":
            original_image = np.argmax(original_image, axis=-1)
        elif label_combination_type == "average":
            original_image = np.argmax(np.round(original_image / number_of_hits), axis=-1)
        else:
            raise ValueError("Unknown combination type")

        # Need to transpose because of different indexing between numpy and simpleitk
        original_image = np.transpose(original_image, transpose_dims)
        return original_image

    def image_array_to_sitk(self, image_array: np.ndarray, input_name: str) -> sitk.Image:
        original_image_direction = self.sample_metadata[input_name]["original_direction"]
        original_image_origin = self.sample_metadata[input_name]["original_origin"]
        original_image_spacing = self.sample_metadata[input_name]["original_spacing"]
        img = sitk.GetImageFromArray(image_array)
        img.SetDirection(original_image_direction)
        img.SetOrigin(original_image_origin)
        img.SetSpacing(original_image_spacing)
        # To ensure proper loading
        img = sitk.Cast(img, sitk.sitkFloat32)

        return img

    def _find_sample_names_from_patch_names(self, data_generator):
        filenames = data_generator.sample_files
        file_locations = np.asarray(data_generator.sample_locations)

        # Get the unique names of the files
        sample_names = np.unique([i_file.split("_patch")[0] for i_file in filenames])

        sample_indices = {}
        for i_sample_name in sample_names:
            sample_indices[i_sample_name] = np.squeeze(
                np.argwhere(
                    [i_sample_name == i_filename.split("_patch")[0] for i_filename in filenames]
                )
            )

        return sample_names, sample_indices

    def get_sample_result_from_patch_results(self, patch_results):
        sample_results = {}
        sample_names = {}
        for i_dataset_name, i_dataset_generator in self.data_generator.items():
            i_patch_results = patch_results[i_dataset_name]
            sample_results[i_dataset_name] = {}
            file_locations = np.asarray(i_dataset_generator.sample_locations)

            sample_names[i_dataset_name], sample_indices = self._find_sample_names_from_patch_names(
                i_dataset_generator
            )

            for i_output_name, i_output_prediction in i_patch_results.items():
                sample_results[i_dataset_name][i_output_name] = []

                if i_output_name in self.image_output_labels:
                    for i_sample_name, i_sample_indices in sample_indices.items():
                        patches_from_sample_results = i_output_prediction[i_sample_indices]
                        sample_results[i_dataset_name][i_output_name].append(
                            self.patches_to_sample_image(
                                i_dataset_generator,
                                file_locations[i_sample_indices],
                                i_output_name,
                                patches_from_sample_results,
                                self.one_hot_outputs[i_output_name],
                                self.config.get_label_combination_type(),
                            )
                        )
                else:
                    for i_sample_name, i_sample_indices in sample_indices.items():
                        patches_from_sample_results = i_output_prediction[i_sample_indices]
                        sample_results[i_dataset_name][i_output_name].append(
                            self.combine_predictions(
                                patches_from_sample_results,
                                self.one_hot_outputs[i_output_name],
                                self.config.get_label_combination_type(),
                            )
                        )
                for i_key, i_value in sample_results[i_dataset_name].items():
                    sample_results[i_dataset_name][i_key] = np.asarray(i_value)

        return sample_names, sample_results

    def get_sample_labels_from_patch_labels(self):
        patch_labels = self.get_real_labels()
        sample_names, sample_labels = self.get_sample_result_from_patch_results(patch_labels)

        return sample_names, sample_labels

    def get_sample_predictions_from_patch_predictions(self):
        patch_predictions = self.predict()
        sample_names, sample_predictions = self.get_sample_result_from_patch_results(
            patch_predictions
        )
        return sample_names, sample_predictions

    @staticmethod
    def one_hot_labels_to_flat_labels(labels: np.ndarray) -> np.ndarray:
        flat_labels = np.argmax(labels, axis=-1)
        flat_labels[labels[..., 0] == -1] = -1
        return flat_labels

    def make_dataframe(self, sample_names, predictions, labels) -> pd.DataFrame:
        df_columns = ["Sample"]
        for i_output_name in self.output_names:
            if i_output_name not in self.image_output_labels:
                df_columns.append("Label_" + i_output_name)
                for i_class in range(self.output_classes[i_output_name]):
                    df_columns.append("Prediction_" + i_output_name + "_class_" + str(i_class))

        results_df = pd.DataFrame(columns=df_columns)
        results_df["Sample"] = sample_names
        for i_output_name, i_output_prediction in predictions.items():
            if i_output_name not in self.image_output_labels:
                # i_output_labels = labels[i_output_name]
                # if labels_one_hot[i_output_name]:
                #     i_output_prediction = self.one_hot_labels_to_flat_labels(i_output_prediction,)
                # i_output_labels = self.one_hot_labels_to_flat_labels(i_output_labels)

                if self.one_hot_outputs[i_output_name]:
                    results_df["Label_" + i_output_name] = self.one_hot_labels_to_flat_labels(
                        labels[i_output_name]
                    )
                else:
                    results_df["Label_" + i_output_name] = labels[i_output_name]
                for i_class in range(self.output_classes[i_output_name]):
                    results_df[
                        "Prediction_" + i_output_name + "_class_" + str(i_class)
                    ] = i_output_prediction[:, i_class]

        return results_df

    def write_image_predictions_to_files(self, sample_names, predictions, labels_one_hot) -> None:
        for i_output_name, i_output_prediction in predictions.items():
            if i_output_name in self.image_output_labels:
                if labels_one_hot is not None and labels_one_hot[i_output_name]:
                    i_output_prediction = self.one_hot_labels_to_flat_labels(i_output_prediction,)

                i_output_prediction_images = [
                    self.image_array_to_sitk(
                        i_sample_output_prediction, self.image_output_labels[i_output_name]
                    )
                    for i_sample_output_prediction in i_output_prediction
                ]

                for i_pred_image, i_sample_name in zip(i_output_prediction_images, sample_names):
                    out_file = os.path.join(
                        self.output_folder, i_sample_name.split(".")[0] + "_prediction.nii.gz"
                    )
                    sitk.WriteImage(i_pred_image, out_file)

    def write_predictions_to_file(self) -> None:
        predictions = self.predict()

        for i_dataset_name, i_dataset_generator in self.data_generator.items():
            if self.config.get_patch_predictions():
                out_file = os.path.join(self.output_folder, i_dataset_name + "_predictions.csv")
                i_prediction = predictions[i_dataset_name]

                results_df = self.make_dataframe(
                    i_dataset_generator.sample_files,
                    i_prediction,
                    self.get_real_labels_of_sample_subset(i_dataset_name),
                )

                results_df.to_csv(out_file, index=False)

                self.write_image_predictions_to_files(
                    i_dataset_generator.sample_files, i_prediction, self.one_hot_outputs,
                )

            if self.config.get_combine_patch_predictions():
                (
                    sample_names,
                    sample_predictions,
                ) = self.get_sample_predictions_from_patch_predictions()
                out_file = os.path.join(
                    self.output_folder, i_dataset_name + "_predictions_combined_patches.csv"
                )

                _, sample_labels = self.get_sample_labels_from_patch_labels()
                results_df = self.make_dataframe(
                    sample_names[i_dataset_name],
                    sample_predictions[i_dataset_name],
                    sample_labels[i_dataset_name],
                )
                results_df.to_csv(out_file, index=False)

                self.write_image_predictions_to_files(
                    sample_names[i_dataset_name], sample_predictions[i_dataset_name], None,
                )

    def make_metric_dataframe(self, metrics: dict) -> pd.DataFrame:
        metrics_df = pd.DataFrame(columns=["Metric"] + self.dataset_names)
        for i_dataset_name in self.dataset_names:
            dataset_metrics = metrics[i_dataset_name]
            metric_names = []
            metric_values = []
            for i_output_name, i_output_metrics in dataset_metrics.items():
                metric_names.extend(
                    [
                        i_output_name + "_" + i_metric_name
                        for i_metric_name in i_output_metrics.keys()
                    ]
                )
                metric_values.extend(list(i_output_metrics.values()))

            metrics_df["Metric"] = metric_names
            metrics_df[i_dataset_name] = metric_values
        return metrics_df

    def write_metrics_to_file(self) -> None:
        if self.config.get_patch_predictions():
            metrics = self.evaluate_metrics()
            out_file = os.path.join(self.output_folder, "metrics.csv")

            metrics_df = self.make_metric_dataframe(metrics)
            metrics_df.to_csv(out_file, index=False)

        if self.config.get_combine_patch_predictions():
            out_file = os.path.join(self.output_folder, "metrics_combined_patches.csv")
            sample_metrics = self.evaluate_sample_metrics()
            sample_metrics_df = self.make_metric_dataframe(sample_metrics)
            sample_metrics_df.to_csv(out_file, index=False)

    def evaluate(self):
        if self.config.get_write_predictions():
            self.write_predictions_to_file()

        if self.config.get_evaluate_metrics():
            self.write_metrics_to_file()

    @classmethod
    def init_from_sys_args(cls, args_in):
        parser = argparse.ArgumentParser(description="Train a CNN")

        parser.add_argument(
            "-c",
            "--config",
            required=True,
            help="The location of the PrognosAIs config file",
            metavar="configuration file",
            dest="config",
            type=str,
        )

        parser.add_argument(
            "-i",
            "--input",
            required=True,
            help="The input directory where the to evaluate samples are located",
            metavar="Input directory",
            dest="input_dir",
            type=str,
        )

        parser.add_argument(
            "-o",
            "--output",
            required=True,
            help="The output directory where to store the results",
            metavar="Output directory",
            dest="output_dir",
            type=str,
        )

        parser.add_argument(
            "-m",
            "--model",
            required=True,
            help="The model file to evaluate",
            metavar="Modefile",
            dest="model_file",
            type=str,
        )

        args = parser.parse_args(args_in)

        return cls(args.model_file, args.input_dir, args.config, args.output_dir)


if __name__ == "__main__":
    evaluator = Evaluator.init_from_sys_args(sys.argv[1:])
    evaluator.evaluate()



from PrognosAIs.Model.Architectures.Architecture import ClassificationNetworkArchitecture
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import Conv3D
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import Dropout
from tensorflow.keras.layers import Flatten
from tensorflow.keras.layers import MaxPooling2D
from tensorflow.keras.layers import MaxPooling3D
from tensorflow.keras.layers import ReLU
from tensorflow.keras.models import Model


class AlexNet_2D(ClassificationNetworkArchitecture):
    padding_type = "valid"

    def create_model(self):
        self.check_minimum_input_size(self.inputs, [67, 67])
        stride_size = self.get_corrected_stride_size(self.inputs, [4, 4], [11, 11])

        conv_1 = Conv2D(filters=96, kernel_size=(11, 11), strides=stride_size)(self.inputs)

        relu_1 = ReLU()(conv_1)

        stride_size = self.get_corrected_stride_size(relu_1, [2, 2], [3, 3])

        pooling_1 = MaxPooling2D(
            pool_size=(3, 3),
            strides=stride_size,
            padding=self.padding_type,
            data_format="channels_last",
        )(relu_1)

        conv_2 = Conv2D(filters=256, kernel_size=(5, 5), strides=(1, 1), padding="same")(pooling_1)

        relu_2 = ReLU()(conv_2)

        stride_size = self.get_corrected_stride_size(relu_2, [2, 2], [3, 3])

        pooling_2 = MaxPooling2D(
            pool_size=(3, 3),
            strides=stride_size,
            padding=self.padding_type,
            data_format="channels_last",
        )(relu_2)

        conv_3 = Conv2D(filters=384, kernel_size=(3, 3), strides=(1, 1), padding="same")(pooling_2)

        relu_3 = ReLU()(conv_3)

        conv_4 = Conv2D(filters=384, kernel_size=(3, 3), strides=(1, 1), padding="same")(relu_3)

        relu_4 = ReLU()(conv_4)

        conv_5 = Conv2D(filters=256, kernel_size=(3, 3), strides=(1, 1), padding="same")(relu_4)

        relu_5 = ReLU()(conv_5)

        stride_size = self.get_corrected_stride_size(relu_5, [2, 2], [3, 3])

        pooling_5 = MaxPooling2D(
            pool_size=(3, 3),
            strides=stride_size,
            padding=self.padding_type,
            data_format="channels_last",
        )(relu_5)

        flatten_1 = Flatten()(pooling_5)

        dense_1 = Dense(
            units=4096,
            activation=None,
            use_bias=True,
            kernel_initializer="glorot_uniform",
            bias_initializer="zeros",
            kernel_regularizer=None,
            bias_regularizer=None,
            activity_regularizer=None,
            kernel_constraint=None,
            bias_constraint=None,
        )(flatten_1)

        dropout_1 = Dropout(0.5)(dense_1)

        dense_2 = Dense(
            units=4096,
            activation=None,
            use_bias=True,
            kernel_initializer="glorot_uniform",
            bias_initializer="zeros",
            kernel_regularizer=None,
            bias_regularizer=None,
            activity_regularizer=None,
            kernel_constraint=None,
            bias_constraint=None,
        )(dropout_1)

        dropout_2 = Dropout(0.5)(dense_2)

        predictions = self.outputs(dropout_2)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model


class AlexNet_3D(ClassificationNetworkArchitecture):
    padding_type = "valid"

    def create_model(self):
        self.check_minimum_input_size(self.inputs, [67, 67, 67])
        stride_size = self.get_corrected_stride_size(self.inputs, [4, 4, 4], [11, 11, 11])

        conv_1 = Conv3D(filters=96, kernel_size=(11, 11, 11), strides=stride_size)(self.inputs)

        relu_1 = ReLU()(conv_1)

        stride_size = self.get_corrected_stride_size(relu_1, [2, 2, 2], [3, 3, 3])

        pooling_1 = MaxPooling3D(
            pool_size=(3, 3, 3),
            strides=stride_size,
            padding=self.padding_type,
            data_format="channels_last",
        )(relu_1)

        conv_2 = Conv3D(filters=256, kernel_size=(5, 5, 5), strides=(1, 1, 1), padding="same")(
            pooling_1
        )

        relu_2 = ReLU()(conv_2)

        stride_size = self.get_corrected_stride_size(relu_2, [2, 2, 2], [3, 3, 3])

        pooling_2 = MaxPooling3D(
            pool_size=(3, 3, 3),
            strides=stride_size,
            padding=self.padding_type,
            data_format="channels_last",
        )(relu_2)

        conv_3 = Conv3D(filters=384, kernel_size=(3, 3, 3), strides=(1, 1, 1), padding="same")(
            pooling_2
        )

        relu_3 = ReLU()(conv_3)

        conv_4 = Conv3D(filters=384, kernel_size=(3, 3, 3), strides=(1, 1, 1), padding="same")(
            relu_3
        )

        relu_4 = ReLU()(conv_4)

        conv_5 = Conv3D(filters=256, kernel_size=(3, 3, 3), strides=(1, 1, 1), padding="same")(
            relu_4
        )

        relu_5 = ReLU()(conv_5)

        stride_size = self.get_corrected_stride_size(relu_5, [2, 2, 2], [3, 3, 3])

        pooling_5 = MaxPooling3D(
            pool_size=(3, 3, 3),
            strides=stride_size,
            padding=self.padding_type,
            data_format="channels_last",
        )(relu_5)

        flatten_1 = Flatten()(pooling_5)

        dense_1 = Dense(
            units=4096,
            activation=None,
            use_bias=True,
            kernel_initializer="glorot_uniform",
            bias_initializer="zeros",
            kernel_regularizer=None,
            bias_regularizer=None,
            activity_regularizer=None,
            kernel_constraint=None,
            bias_constraint=None,
        )(flatten_1)

        dropout_1 = Dropout(0.5)(dense_1)

        dense_2 = Dense(
            units=4096,
            activation=None,
            use_bias=True,
            kernel_initializer="glorot_uniform",
            bias_initializer="zeros",
            kernel_regularizer=None,
            bias_regularizer=None,
            activity_regularizer=None,
            kernel_constraint=None,
            bias_constraint=None,
        )(dropout_1)

        dropout_2 = Dropout(0.5)(dense_2)

        predictions = self.outputs(dropout_2)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model



import copy
import importlib
import os

import tensorflow.keras.callbacks
import tensorflow.keras.losses
import tensorflow.keras.metrics
import tensorflow.keras.optimizers

import PrognosAIs.Model.Callbacks
import PrognosAIs.Model.Losses
import PrognosAIs.Model.Metrics

from PrognosAIs.IO import utils as IO_utils


class StandardParser:
    def __init__(self, config: dict, module_paths: list):
        self.config = config
        self.module_paths = module_paths

    def parse_settings(self):
        if self.config is not None:
            if "name" in self.config:
                parsed_settings = self._initiate_class(self.config["name"], self.config["settings"])
            else:
                parsed_settings = {}

                for key, value in self.config.items():
                    if "name" in value:
                        parsed_settings[key] = self._initiate_class(
                            value["name"], value["settings"]
                        )
                    else:
                        settings = []
                        for key_deep, value_deep in value.items():
                            settings.append(
                                self._initiate_class(value_deep["name"], value_deep["settings"])
                            )
                        parsed_settings[key] = settings
        else:
            parsed_settings = None

        return parsed_settings

    def _initiate_class(self, class_name, class_settings):
        class_function = self.get_class(class_name)

        if class_settings is None:
            initiated_class = class_function()
        else:
            initiated_class = class_function(**class_settings)

        return initiated_class

    def get_class(self, class_name):
        class_function = None
        module_index = 0
        while class_function is None and module_index < len(self.module_paths):
            module_path = self.module_paths[module_index]
            if isinstance(module_path, str):
                module_path = IO_utils.load_module_from_file(module_path)
            class_function = getattr(module_path, class_name, None)

            module_index += 1

        if class_function is None:
            raise ValueError("Requested class {} not found!".format(class_name))

        return class_function


class LossParser(StandardParser):
    def __init__(self, loss_settings: dict, class_weights: dict = None, module_paths=None):
        """
        Parse loss settings to actual losses

        Args:
            loss_settings: Settings for the losses

        Returns:
            None
        """

        # TODO remove the need for a "settings" index if there is no settings
        # Index just need to default to None.
        # Also get rid of all the "vague" keywords that are then in the config

        self.class_weights = class_weights
        self.module_paths = [tensorflow.keras.losses, PrognosAIs.Model.Losses]
        if module_paths is not None:
            if isinstance(module_paths, list):
                self.module_paths.extend(module_paths)
            else:
                self.module_paths.append(module_paths)

        if "name" in loss_settings and class_weights is not None:
            loss_settings["settings"]["class_weight"] = self.class_weights
        elif "name" not in loss_settings:
            for key, value in loss_settings.items():
                if (
                    value["settings"] is not None
                    and class_weights is not None
                    and key in self.class_weights.keys()
                ):
                    value["settings"]["class_weight"] = self.class_weights[key]

        super().__init__(loss_settings, self.module_paths)

    def get_losses(self):
        return self.parse_settings()


class OptimizerParser(StandardParser):
    def __init__(self, optimizer_settings: dict, module_paths=None) -> None:
        """
        Interfacing class to easily get a tf.keras.optimizers optimizer

        Args:
            optimizer_settings: Arguments to be passed to the optimizer

        Returns:
            None
        """

        self.module_paths = [tensorflow.keras.optimizers]
        if module_paths is not None:
            if isinstance(module_paths, list):
                self.module_paths.extend(module_paths)
            else:
                self.module_paths.append(module_paths)

        super().__init__(optimizer_settings, self.module_paths)

        return

    def get_optimizer(self):
        return self.parse_settings()


class CallbackParser(StandardParser):
    def __init__(
        self, callback_settings: dict, root_path: str = None, module_paths=None, save_name=None
    ):
        """
        Parse callback settings to actual callbacks

        Args:
            callback_settings: Settings for the callbacks

        Returns:
            None
        """
        self.module_paths = [tensorflow.keras.callbacks, PrognosAIs.Model.Callbacks]
        self.save_name = save_name
        if module_paths is not None:
            if isinstance(module_paths, list):
                self.module_paths.extend(module_paths)
            else:
                self.module_paths.append(module_paths)

        super().__init__(callback_settings, self.module_paths)

        if root_path is not None:
            self.config = self.replace_root_path(self.config, root_path)
        return

    def replace_root_path(self, settings, root_path):
        for key, value in settings.items():
            if type(value) == dict:
                settings[key] = self.replace_root_path(value, root_path)
            else:
                # TODO  this gives error, for example with TensorBoard which has profile_batch option
                # (because file is in profile)
                if "file" in key:
                    settings[key] = os.path.join(root_path, value)
                    if self.save_name is not None:
                        settings[key] = settings[key].format(savename=self.save_name)

        return settings

    def get_callbacks(self):
        # Need to make sure that CSVLogger is the last so that everything is properly stored
        parsed_callbacks = list(self.parse_settings().values())
        out_parsed_callbacks = []

        logger_index = -1
        for i_i_parsed_callback, i_parsed_callback in enumerate(parsed_callbacks):
            if type(i_parsed_callback) == tensorflow.keras.callbacks.CSVLogger:
                logger_index = i_i_parsed_callback
            else:
                out_parsed_callbacks.append(i_parsed_callback)

        if logger_index != -1:
            out_parsed_callbacks.append(parsed_callbacks[logger_index])

        return out_parsed_callbacks


class MetricParser(StandardParser):
    def __init__(self, metric_settings: dict, label_names: list = None, module_paths=None) -> None:
        """
        Parse metrics settings to actual metrics

        Args:
            loss_settings: Settings for the losses
        """

        self.module_paths = [tensorflow.keras.metrics, PrognosAIs.Model.Metrics]
        if module_paths is not None:
            if isinstance(module_paths, list):
                self.module_paths.extend(module_paths)
            else:
                self.module_paths.append(module_paths)

        super().__init__(metric_settings, self.module_paths)
        if label_names is not None:
            self.label_names = label_names
        else:
            self.label_names = []

    def get_metrics(self):
        parsed_callbacks = self.parse_settings()

        if not isinstance(parsed_callbacks, dict) and parsed_callbacks is not None:
            parsed_callbacks = [parsed_callbacks]
        elif (
            parsed_callbacks is not None
            and self.label_names != []
            and sorted(parsed_callbacks.keys()) != sorted(self.label_names)
        ):
            parsed_callbacks = list(parsed_callbacks.values())
        elif parsed_callbacks is None:
            parsed_callbacks = []
        return parsed_callbacks

    def convert_metrics_list_to_dict(self, metrics: list) -> dict:
        if isinstance(metrics, list):
            new_metrics = {}

            for i_label_name in self.label_names:
                new_metrics[i_label_name] = []
                other_label_names = copy.deepcopy(self.label_names)
                other_label_names.remove(i_label_name)

                for i_metric in metrics:
                    i_metric_name = i_metric.name
                    i_metric_base_name = i_metric.__class__().name
                    i_metric_label_name = i_metric_name.split(i_metric_base_name)[0]
                    # No output specific metrics, append all metrics to all outputs
                    if "_" not in i_metric_label_name:
                        new_metrics[i_label_name].append(i_metric)
                    else:
                        # Need to cut off the last underscore to get the label name
                        i_metric_label_name = i_metric_label_name[:-1]
                        if i_metric_label_name == i_label_name:
                            new_metrics[i_label_name].append(i_metric)
        elif isinstance(metrics, dict):
            for key, value in metrics.items():
                if not isinstance(value, list):
                    metrics[key] = [value]

            new_metrics = metrics
        else:
            new_metrics = metrics
        return new_metrics



from typing import Any, Tuple, Union
import logging

import h5py
import numpy as np
import tensorflow as tf
import tensorflow_io as tfio
import scipy

import PrognosAIs.Constants
import PrognosAIs.IO.utils as IO_utils


class HDF5Generator:
    def __init__(
        self,
        root_folder: str,
        batch_size: int = 16,
        shuffle: bool = False,
        max_steps: int = -1,
        drop_batch_remainder: bool = True,
        labels_only: bool = False,
    ) -> None:
        """
        Generate data from HDF5 files to be used in a TensorFlow pipeline.

        This generator loads sample data from HDF5 files, and does this efficiently
        making us of TensorFlow dataset functions.
        The inputs and outputs are dict, which allows for easy us in a
        multi-input and/or multi-output model

        Args:
            root_folder (str): Folder in which the HDF5 files are stored
            batch_size (int, optional): Batch size of the generator. Defaults to 16.
            shuffle (bool, optional): Whether datset should be shuffled. Defaults to False.
            data_augmentation (bool, optional): Whether data augmentation should be applied.
                Defaults to False.
            augmentation_factor (int, optional): Number of times dataset should be repeated for augmentation.
                Defaults to 5.
            augmentation_settings (dict, optional): Setting for the data augmenation. Defaults to None.
            max_steps (int, optional): Maximum number of (iteration) steps to provide.
                Defaults to -1, in which case all samples are provied.
            drop_batch_remainder (bool, optional): Whether to drop the remainder of the batch if it does not fit perfectly.
                Defaults to True.
            labels_only (bool, optional): Whether to only provide labels. Defaults to False.
            feature_index (str, optional): Name of the feature group in the HDF5 file. Defaults to "sample".
            label_index (str, optional): Name of the label group in the HDF5 file. Defaults to "label".
        """
        self.cache_in_memory = False
        self.TF_dataset = None
        self.augmentation_factor = 1
        self.augmentation_settings = {}
        self.augmentors = {}
        self.data_augmentation = False
        self.feature_index = PrognosAIs.Constants.FEATURE_INDEX
        self.label_index = PrognosAIs.Constants.LABEL_INDEX
        self.repeat = False
        self.shard = False
        self.n_workers = 1
        self.worker_index = 0

        self.sample_locations = IO_utils.find_files_with_extension(
            root_folder, PrognosAIs.Constants.HDF5_EXTENSION,
        )

        self.sample_files = [
            IO_utils.get_file_name_from_full_path(i_sample_location)
            for i_sample_location in self.sample_locations
        ]

        self.example_sample_file = self.sample_locations[0]
        self.max_steps = max_steps
        self.memory_size = IO_utils.get_dir_size(root_folder)

        self.batch_size = batch_size
        self.shuffle = shuffle
        self.N_samples = len(self.sample_locations)
        self.drop_batch_remainder = drop_batch_remainder

        self.labels_only = labels_only
        self.dataset_attributes = self.get_all_dataset_attributes()

        self.hdf_dataset_names = self.get_dataset_names()
        self.feature_dataset_names = [
            i_ds_name for i_ds_name in self.hdf_dataset_names if self.feature_index in i_ds_name
        ]
        self.label_dataset_names = [
            i_ds_name for i_ds_name in self.hdf_dataset_names if self.label_index in i_ds_name
        ]
        self.feature_names = [
            i_ds_name.split(PrognosAIs.Constants.HDF_SEPARATOR)[-1]
            for i_ds_name in self.feature_dataset_names
        ]
        self.label_names = [
            i_ds_name.split(PrognosAIs.Constants.HDF_SEPARATOR)[-1]
            for i_ds_name in self.label_dataset_names
        ]

        self.spec = self.get_spec()
        self.possible_steps = int(np.floor(self.N_samples / self.batch_size))
        self.labels_are_one_hot = self.get_labels_are_one_hot()

        if self.max_steps != -1:
            self.steps = np.minimum(self.possible_steps, self.max_steps)
        else:
            self.steps = self.possible_steps

        if self.labels_only:
            self.loader = self.label_loader
        else:
            self.loader = self.features_and_labels_loader

    def get_spec(self) -> dict:
        """
        Get the TensorSpec for all input features.

        Returns:
            dict: Maps the name of each input feature to the TensorSpec of the input.
        """
        spec = {}
        for i_ds_name in self.hdf_dataset_names:
            dtype = self.get_dataset_attribute(i_ds_name, "dtype")
            shape = self.get_dataset_attribute(i_ds_name, "shape")
            spec[i_ds_name] = tf.TensorSpec(shape, dtype=dtype, name=i_ds_name)

        return spec

    def get_dataset_names(self) -> list:
        """
        Get the names of all datasets in the sample.

        Returns:
            list: Dataset names in the sample
        """
        with h5py.File(self.example_sample_file, "r") as example_5py:
            ds_names = self._get_dataset_names(example_5py)
        return ds_names

    def _get_dataset_names(self, h5py_object: Union[h5py.File, h5py.Dataset, h5py.Group]) -> list:
        """
        Run through all groups and dataset to get the names.

        Args:
            h5py_object (Union[h5py.File, h5py.Dataset, h5py.Group]): Object for which to return
                the dataset names

        Returns:
            list: Dataset names in object
        """
        if isinstance(h5py_object, h5py.Dataset):
            dataset_names = [h5py_object.name]
        else:
            dataset_names = []
            for i_key in h5py_object.keys():
                dataset_names.extend(self._get_dataset_names(h5py_object.get(i_key)))
        return dataset_names

    def get_all_dataset_attributes(self, sample_file: str = None) -> dict:
        """
        Get the attributes of the features and labels stored in the file.

        Returns:
            dict: Mapping of the feature/label name to its attributes
        """
        if sample_file is None:
            sample_file = self.example_sample_file
        with h5py.File(sample_file, "r") as hdf5_example:
            ds_attributes = self._get_all_dataset_attributes(hdf5_example)

        return ds_attributes

    def _get_all_dataset_attributes(
        self, h5py_object: Union[h5py.File, h5py.Dataset, h5py.Group],
    ) -> dict:
        """
        Run through al groups and dataset to get the attributes.

        Args:
            h5py_object (Union[h5py.File, h5py.Dataset, h5py.Group]): Object for which to return
                the attributes

        Returns:
            dict: Mapping between feature/label name and its attributes
        """
        if isinstance(h5py_object, h5py.Dataset):
            ds_attributes = dict(h5py_object.attrs)
            ds_attributes["shape"] = h5py_object.shape
            ds_attributes["dtype"] = h5py_object.dtype
            return {h5py_object.name: ds_attributes}
        else:
            ds_attributes = {h5py_object.name: dict(h5py_object.attrs)}
            for i_key in h5py_object.keys():
                ds_attributes.update(self._get_all_dataset_attributes(h5py_object.get(i_key)))
            return ds_attributes

    def get_dataset_attribute(self, dataset_name: str, attribute_name: str) -> Any:
        """
        Get the attribute of a specific dataset

        Args:
            dataset_name (str): Name of dataset for which to get the attribute
            attribute_name (str): Name of attribute to get

        Returns:
            Any: The value of the attribute
        """
        return self.dataset_attributes[dataset_name][attribute_name]

    def get_feature_attribute(self, attribute_name: str) -> dict:
        """
        Get a specific attribute for all features.

        Args:
            attribute_name (str): Name of attribute to get

        Returns:
            dict: Mapping between feature names and the attribute value
        """
        attribute = {}
        for i_feature_name, i_feature_dataset_name in zip(
            self.feature_names, self.feature_dataset_names,
        ):
            attribute[i_feature_name] = self.get_dataset_attribute(
                i_feature_dataset_name, attribute_name,
            )

        return attribute

    def get_label_attribute(self, attribute_name: str) -> dict:
        """
        Get a specific attribute for all labels.

        Args:
            attribute_name (str): Name of attribute to get

        Returns:
            dict: Mapping between label names and the attribute value
        """
        attribute = {}
        for i_label_name, i_label_dataset_name in zip(self.label_names, self.label_dataset_names):
            attribute[i_label_name] = self.get_dataset_attribute(
                i_label_dataset_name, attribute_name,
            )

        return attribute

    def get_feature_metadata(self) -> dict:
        """
        Get all metadata of all features.

        Returns:
            dict: The metadata of all features
        """
        feature_metadata = {}

        for i_feature_name, i_feature_dataset_name in zip(
            self.feature_names, self.feature_dataset_names,
        ):
            feature_metadata[i_feature_name] = self.dataset_attributes[i_feature_dataset_name]

        return feature_metadata

    def get_feature_metadata_from_sample(self, sample_location: str) -> dict:
        """
        Get the feature metadata of a specific sample.

        Args:
            sample_location (str): The file location of the sample

        Returns:
            dict: The feature metadata of the sample
        """
        ds_attributes = self.get_all_dataset_attributes(sample_location)
        feature_metadata = {}
        for i_feature_name, i_feature_dataset_name in zip(
            self.feature_names, self.feature_dataset_names,
        ):
            feature_metadata[i_feature_name] = ds_attributes[i_feature_dataset_name]
        return feature_metadata

    def get_number_of_classes(self) -> dict:
        """
        Get the number of output classes.

        Returns:
            dict: Number of output classes for each label
        """
        return self.get_label_attribute("N_classes")

    def get_feature_dimensionality(self) -> dict:
        """
        Get the dimensionality of each feature.

        Returns:
            dict: Dimensionality of each feature
        """
        return self.get_feature_attribute("dimensionality")

    def get_feature_size(self) -> dict:
        """
        Get the size of each feature.

        The size only of the feature does not take into account the number of channels
        and only represents the size of an individual channel of the feature.

        Returns:
            dict: Size of each feature
        """
        return self.get_feature_attribute("size")

    def get_feature_shape(self) -> dict:
        """
        Get the shape of each feature.

        Returns:
            dict: Shape of each feature
        """
        return self.get_feature_attribute("shape")

    def get_number_of_channels(self) -> dict:
        """
        Get the number of feature channels.

        Returns:
            dict: Number of channels for each feature
        """
        return self.get_feature_attribute("N_channels")

    def get_labels_are_one_hot(self) -> dict:
        """
        Get whether labels are one-hot encoded.

        Returns:
            dict: One-hot encoding status of each label
        """
        return self.get_label_attribute("one_hot")

    def setup_augmentation(
        self, augmentation_factor: int = 1, augmentation_settings: dict = {},
    ) -> None:
        """
        Set up data augmentation in the generator.

        Args:
            augmentation_factor (int): Repeat dataset this many times in augmentation. Defaults to 1.
            augmentation_settings (dict): Setting to parse to augmentation instance. Defaults to {}.
        """

        logging.info(
            (
                "Setting up data augmentation with the following settings:\n"
                "Augmentation factor: {aug_fac}\n"
                "Augmentation settings: {aug_set}"
            ).format(aug_fac=augmentation_factor, aug_set=augmentation_settings)
        )
        self.augmentation_factor = augmentation_factor
        self.augmentation_settings = augmentation_settings
        self.data_augmentation = True
        for i_feature_ds_name, i_feature_name in zip(
            self.feature_dataset_names, self.feature_names,
        ):
            self.augmentors[i_feature_name] = Augmentor(
                self.spec[i_feature_ds_name], **self.augmentation_settings,
            )

        for i_label_ds_name, i_label_name in zip(self.label_dataset_names, self.label_names,):
            self.augmentors[i_label_name] = Augmentor(
                self.spec[i_label_ds_name], **self.augmentation_settings,
            )

    def features_and_labels_loader(
        self, sample_location: tf.Tensor
    ) -> Tuple[dict, dict, tf.Tensor]:
        """
        Load the features and labels from a hdf5 file to be used in a TensorFlow dataset pipeline.

        This loader loads the features and labels from a hdf5 file using TensorFlowIO.
        The outputs are therefor directly cast to tensor and can be used in a TensorFlow graph.
        All features and labels from the file are loaded, and a dict is returned mapping
        the name of each feature and label to its respective value

        Args:
            sample_location (tf.Tensor): Location of the sample file

        Returns:
            Tuple[dict, dict]: The features (first output) and labels (second output) loaded
                from the sample.
        """
        # tf.strings.split doesnt work because we can then not use the
        # value of that tensor as an index for the dict.
        # Works in eager execution but not in graph execution.
        # if that works in the future, this can be replace as it is probably faster.
        # Maybe we can do something by loading each dataset, concatenating and then returning
        # the name as well.
        # This could be done using tfio.IODataset instead of IOTensor, but might
        # be complicated then with concatenation and such
        loaded_hdf5 = tfio.IOTensor.from_hdf5(sample_location, spec=self.spec)
        features = self.load_features(loaded_hdf5)
        labels = self.load_labels(loaded_hdf5)

        return features, labels

    def label_loader(self, sample_location: tf.Tensor) -> dict:
        """
        Load the labels from a hdf5 sample file.

        This loader only loads the labels, instead of the features and labels
        as done by features_and_labels_loader

        Args:
            sample_location (tf.Tensor): Location of the sample file

        Returns:
            dict: Labels loaded from the sample file
        """
        loaded_hdf5 = tfio.IOTensor.from_hdf5(sample_location, spec=self.spec)
        return self.load_labels(loaded_hdf5)

    def feature_loader(self, sample_location: tf.Tensor) -> dict:
        """
        Load the features from a hdf5 sample file.

        This loader only loads the labels, instead of the features and labels
        as done by features_and_labels_loader

        Args:
            sample_location (tf.Tensor): Location of the sample file

        Returns:
            dict: Features loaded from the sample file
        """
        loaded_hdf5 = tfio.IOTensor.from_hdf5(sample_location, spec=self.spec)
        return self.load_features(loaded_hdf5)

    def load_features(self, loaded_hdf5: tfio.IOTensor) -> dict:
        """
        Load the features from a HDF5 tensor.

        Args:
            loaded_hdf5 (tfio.IOTensor): Tensor from which to load features

        Returns:
            dict: Mapping between feature names and features
        """
        features = {}
        for i_feature_ds_name, i_feature_name in zip(
            self.feature_dataset_names, self.feature_names,
        ):
            features[i_feature_name] = loaded_hdf5(i_feature_ds_name).to_tensor()

        return features

    def load_labels(self, loaded_hdf5: tfio.IOTensor) -> dict:
        """
        Load the labels from a HDF5 tensor.

        Args:
            loaded_hdf5 (tfio.IOTensor): Tensor from which to load labels

        Returns:
            dict: Mapping between label names and labels
        """
        labels = {}
        for i_label_ds_name, i_label_name in zip(self.label_dataset_names, self.label_names):
            labels[i_label_name] = loaded_hdf5(i_label_ds_name).to_tensor()
        return labels

    def apply_augmentation(self, features: dict, labels: dict) -> Tuple[dict, dict]:
        seed = tf.random.uniform([2], 0, 10000000, dtype=tf.dtypes.int32)
        for i_key, i_value in features.items():
            features[i_key] = self.augmentors[i_key].augment_sample(i_value, seed)

        for i_key, i_value in labels.items():
            if i_key == "MASK":
                labels[i_key] = self.augmentors[i_key].augment_sample(i_value, seed, True)

        return features, labels

    def fits_in_memory(self, used_memory: int = 0):
        ds_size = self.memory_size * PrognosAIs.Constants.MEM_SAFETY_FACTOR
        return ds_size <= IO_utils.get_available_ram(used_memory)

    def setup_caching(
        self, cache_in_memory: Union[bool, str] = PrognosAIs.Constants.AUTO, used_memory: int = 0,
    ) -> None:
        """
        Set up caching of the dataset in RAM.

        Args:
            cache_in_memory (Union[bool, str]): Whether dataset should be cached in memory.
                Defaults to PrognosAIs.Constants.AUTO, in which case the dataset will be
                cached in memory if it fits, otherwise it will not be cached
            used_memory (int): Amount of RAM (in bytes) that is already being used. Defaults to 0.

        Raises:
            ValueError: If an unknown cache setting is requested
        """
        if cache_in_memory == PrognosAIs.Constants.AUTO and self.fits_in_memory(used_memory):
            self.cache_in_memory = True
        elif isinstance(cache_in_memory, bool):
            self.cache_in_memory = cache_in_memory
        elif cache_in_memory not in [PrognosAIs.Constants.AUTO, True, False]:
            err_msg = (
                "Unknown cache in memory setting {cach_set}"
                ", should be either True, False or {auto}"
            ).format(cach_set=cache_in_memory, auto=PrognosAIs.Constants.AUTO,)
            raise ValueError(err_msg)

    def setup_sharding(self, n_workers: int, worker_index: int) -> None:
        """
        Shard the dataset according to the number of workers and worker index

        Args:
            n_workers (int): number of workers
            worker_index (int): worker index
        """
        self.n_workers = n_workers
        self.worker_index = worker_index
        self.shard = True

    def setup_caching_shuffling_steps(self, dataset: tf.data.Dataset) -> tf.data.Dataset:
        """
        Set-up caching, shuffling and the iteration step in the dataset pipeline.

        This function helps to ensure that caching, shuffling and step limiting is done properly
        and efficiently, no matter where in the dataset pipeline it is included.

        Args:
            dataset (tf.data.Dataset): Datset for which to include the steps

        Returns:
            tf.data.Dataset: Datset with caching, shuffling and iteration steps included
        """
        dataset = dataset.cache()
        # Lets go over the dataset to make sure that we cached everything
        # Otherwise it might be that in the first iteration we do not full iterate
        # Over the dataset, making our cache useless
        logging.info("Caching dataset")
        for _ in dataset:
            pass
        logging.info("Done caching dataset")

        if self.data_augmentation:
            dataset = dataset.repeat(self.augmentation_factor)

        if self.shuffle:
            dataset = dataset.shuffle(
                self.N_samples * self.augmentation_factor, reshuffle_each_iteration=True
            )

        if self.steps < self.possible_steps:
            dataset = dataset.take(self.steps * self.batch_size)

        return dataset

    def get_tf_dataset(
        self, num_parallel_calls: int = tf.data.experimental.AUTOTUNE,
    ) -> tf.data.Dataset:
        """
        Construct a TensorFlow dataset.

        The dataset is constructed based on the settings supplied to the DataGenerator.
        The dataset can then directly be used to train or evaluate a TensorFlow model

        Args:
            num_parallel_calls (int): Number of parallel process to use.
                Defaults to tf.data.experimental.AUTOTUNE.

        Returns:
            tf.data.Dataset: The constructed dataset
        """
        if self.TF_dataset is None:
            dataset = tf.data.Dataset.from_tensor_slices(self.sample_locations)

            if self.shard:
                dataset = dataset.shard(self.n_workers, self.worker_index)

            # If we dont cache the whole dataset in memory, we just cache the
            # file names, and perform as many steps before the data loading
            # as possible to reduce the loading times
            if not self.cache_in_memory:
                dataset = self.setup_caching_shuffling_steps(dataset)

            dataset = dataset.map(self.loader, num_parallel_calls=num_parallel_calls)

            if self.cache_in_memory:
                dataset = self.setup_caching_shuffling_steps(dataset)

            if self.data_augmentation:
                dataset = dataset.map(
                    self.apply_augmentation, num_parallel_calls=num_parallel_calls,
                )

            dataset = dataset.batch(self.batch_size, drop_remainder=self.drop_batch_remainder)

            if self.repeat:
                dataset = dataset.repeat()

            dataset = dataset.prefetch(num_parallel_calls)
            self.TF_dataset = dataset
        return self.TF_dataset

    def get_numpy_iterator(self) -> np.nditer:
        """
        Construct a numpy iterator instead of TensorFlow dataset.

        The numpy iterator will provide exactly the same data as the TensorFlow dataset.
        However, it might be easier to inspect the data when using a numpy iterator
        instead of a TensorFlow dataset

        Returns:
            np.nditer: The dataset
        """
        dataset = self.get_tf_dataset()

        return dataset.as_numpy_iterator()


class Augmentor(object):
    """
    Augmentor
    """

    def __init__(
        self,
        example_sample: tf.Tensor,
        brightness_probability: float = 0,
        brightness_delta: float = 0,
        contrast_probability: float = 0,
        contrast_min_factor: float = 1,
        contrast_max_factor: float = 1,
        flip_probability: float = 0,
        to_flip_axis: Union[int, list] = 0,
        crop_probability: float = 0,
        crop_size: list = None,
        rotate_probability: float = 0,
        max_rotate_angle: float = 0,
        to_rotate_axis: Union[int, list] = 0,
    ) -> None:
        """
        Augmentor to randomly augment the features of a sample.

        Args:
            example_sample (tf.Tensor): Example sample from which settings for augmentation
                will be derived
            brightness_probability (float, optional): Probability of augmenting brightness. Defaults to 0.
            brightness_delta (float, optional): Brightness will be adjusted with value from -delta to delta.
                Defaults to 0.
            contrast_probability (float, optional): Probability of augmenting contrast.
                Defaults to 0.
            contrast_min_factor (float, optional): Minimum contrast adjustment factor.
                Defaults to 1.
            contrast_max_factor (float, optional): Maximum contrast adjustment factor.
                Defaults to 1.
            flip_probability (float, optional): Probability of a random flip. Defaults to 0.
            to_flip_axis (Union[int, list], optional): Axis to flip the feature over. Defaults to 0.
            crop_probability (float, optional): Probability of cropping the feature. Defaults to 0.
            crop_size (list, optional): Size to crop the feature to. Defaults to None.
        """

        self.sample_size = example_sample.shape
        # Minus 1 because last dimension is channel and we dont want to augment
        # those (they will be augmented all the same, not individually)
        self.n_dim = len(example_sample.shape) - 1
        self.to_reduce_axis = tf.range(0, self.n_dim, 1)

        self.brightness_probability = brightness_probability
        self.brightness_delta = brightness_delta

        self.contrast_probability = contrast_probability
        self.contrast_min = contrast_min_factor
        self.contrast_max = contrast_max_factor

        self.flip_probability = flip_probability
        if isinstance(to_flip_axis, list):
            self.to_flip_axis = to_flip_axis
        elif isinstance(to_flip_axis, int):
            self.to_flip_axis = [to_flip_axis]

        self.crop_probability = crop_probability
        if crop_size is None:
            self.crop_size = [0, 0, 0]
        else:
            self.crop_size = crop_size

        self.rotate_probability = rotate_probability
        self.max_rotate_angle = max_rotate_angle
        if isinstance(to_rotate_axis, list):
            self.to_rotate_axis = to_rotate_axis
        else:
            self.to_rotate_axis = [to_rotate_axis]

    def get_seed(self) -> tf.Tensor:
        """
        Get a random seed that can be used to make other operation repeatable.

        Returns:
            tf.Tensor: The seed
        """
        return tf.random.uniform([], 0, tf.dtypes.int32.max - 1, dtype=tf.dtypes.int32)

    def apply_augmentation(self, augmentation_probability: float, seed: tf.Tensor = None) -> bool:
        """
        Whether the the augmentation step should be applied based on the probability.

        Args:
            augmentation_probability (float): The probability with which the step should be applied
            seed (tf.Tensor): Seed to make operation repeatable. Defaults to None.

        Returns:
            bool: Whether the step should be applied
        """
        if seed is not None and isinstance(seed, tf.Tensor):
            return (
                tf.random.stateless_uniform([], seed, 0, 1, dtype=tf.float32)
                < augmentation_probability
            )
        else:
            return tf.random.uniform([], 0, 1, dtype=tf.float32) < augmentation_probability

    def random_brightness(self, sample: tf.Tensor, seed: tf.Tensor = None) -> tf.Tensor:
        """
        Randomly adjusts the brightness of a sample.

        Brightness is adjusted by a constact factor over the whole image, drawn from
        a distribution between -delta and delta as set during the initialization of
        the augmentator.

        Args:
            sample (tf.Tensor): Sample for which to adjust brightness.
            seed (tf.Tensor): Seed to make operation repeatable. Defaults to None.

        Returns:
            tf.Tensor: The augmented sample.
        """
        if self.apply_augmentation(self.brightness_probability, seed):
            sample += tf.random.stateless_uniform(
                [], seed, -self.brightness_delta, self.brightness_delta, dtype=sample.dtype,
            )
        return sample

    def random_contrast(self, sample: tf.Tensor, seed: tf.Tensor = None) -> tf.Tensor:
        """
        Randomly adjust the contrast of a sample.

        The contrast is adjusted by keeping the mean of the sample the same as
        for the original sample, and squeezing or expending the distribution of
        the intensities around the mean. The amount of squeezing or expanding is
        randomly drawn from the minimum and maximum contrast set during initialization.

        Args:
            sample (tf.Tensor): Sample for which to adjust contrast
            seed (tf.Tensor): Seed to make operation repeatable. Defaults to None.

        Returns:
            tf.Tensor: The augmented sample
        """
        if self.apply_augmentation(self.contrast_probability, seed):
            contrast_factor = tf.random.stateless_uniform(
                [], seed, self.contrast_min, self.contrast_max, dtype=sample.dtype
            )
            sample_mean = tf.math.reduce_mean(sample, axis=self.to_reduce_axis)
            sample = (sample - sample_mean) * contrast_factor + sample_mean

        return sample

    def random_flipping(self, sample: tf.Tensor, seed: tf.Tensor = None) -> tf.Tensor:
        """
        Randomly flip the sample along one or multiple axis.

        Args:
            sample (tf.Tensor): Sample for which to apply flipping
            seed (tf.Tensor): Seed to make operation repeatable. Defaults to None.

        Returns:
            tf.Tensor: The augmented sample
        """
        for i_flip_axis in self.to_flip_axis:
            if self.apply_augmentation(self.flip_probability, seed + i_flip_axis):
                sample = tf.reverse(sample, [i_flip_axis])
        return sample

    def random_cropping(self, sample: tf.Tensor, seed: tf.Tensor = None) -> tf.Tensor:
        """
        Randomly crop a part of the sample.

        The crop will have the size of the crop size defined upon initialization of the augmentator.
        The crop will happen for all channels in the same way, but will not crop out channels.
        The location of the crop will be randomly drawn from throughout the whole image.

        Args:
            sample (tf.Tensor): The sample to be cropped
            seed (tf.Tensor): Seed to make operation repeatable. Defaults to None.

        Returns:
            tf.Tensor: The augmented sample
        """
        if self.apply_augmentation(self.crop_probability, seed):
            crop_start = []

            for i_dim in range(self.n_dim):
                dim_seed = seed + i_dim
                crop_start.append(
                    tf.random.stateless_uniform(
                        [],
                        dim_seed,
                        0,
                        tf.shape(sample)[i_dim] - self.crop_size[i_dim],
                        dtype=tf.int32,
                    ),
                )

            # Need to make sure we do not crop channels
            crop_start.append(tf.constant(0))
            crop_size = self.crop_size + [tf.shape(sample)[-1]]

            sample = tf.slice(sample, crop_start, crop_size)

        return sample

    def _random_rotate(self, feature, seed, to_rotate_axis, interpolation_order):
        seed = seed + tf.cast(to_rotate_axis, seed.dtype)
        all_axis = tf.range(0, tf.rank(feature) - 1)
        np.random.seed(seed)
        angle = np.random.uniform(-self.max_rotate_angle, self.max_rotate_angle)

        rotation_axis = np.setdiff1d(all_axis, to_rotate_axis)
        feature = scipy.ndimage.rotate(
            feature, angle, reshape=False, order=interpolation_order, axes=rotation_axis
        )
        return feature

    def _rotate(self, feature, angle, interpolation_order, axis):
        return scipy.ndimage.rotate(
            feature, angle, reshape=False, order=interpolation_order, axes=axis,
        )

    def random_rotate(
        self, feature: tf.Tensor, seed: tf.Tensor = None, interpolation_order: int = 3
    ) -> tf.Tensor:
        all_axis = tf.range(0, tf.rank(feature) - 1)
        for i_to_rotate_axis in self.to_rotate_axis:
            axis_seed = seed + i_to_rotate_axis
            if self.apply_augmentation(self.rotate_probability, axis_seed):
                angle = tf.random.stateless_uniform(
                    [], axis_seed, -self.max_rotate_angle, self.max_rotate_angle, dtype=tf.float32,
                )
                rotation_axis, _ = tf.compat.v1.setdiff1d(all_axis, tf.constant([i_to_rotate_axis]))
                feature = tf.numpy_function(
                    self._rotate,
                    [feature, angle, interpolation_order, rotation_axis],
                    feature.dtype,
                )
        return feature

    def pad_to_original_size(self, sample: tf.Tensor) -> tf.Tensor:
        """
        Pad back a (potentially) augmented sample to its original size.

        Args:
            sample (tf.Tensor): The sample to pad

        Returns:
            tf.Tensor: The padded sample with the same size as before any augmentation steps
        """
        sample_size = tf.shape(sample)

        required_paddings = [
            [tf.math.ceil((m - sample_size[i]) / 2), tf.math.floor((m - sample_size[i]) / 2)]
            for (i, m) in enumerate(self.sample_size)
        ]

        return tf.pad(sample, required_paddings)

    def augment_sample(self, sample: tf.Tensor, seed=None, is_mask=False) -> tf.Tensor:
        """
        Apply random augmentations to the sample based on the config.

        Args:
            sample (tf.Tensor): sample to be augmented

        Returns:
            tf.Tensor: augmented sample
        """
        if not is_mask:
            sample = self.random_brightness(sample, seed)
            sample = self.random_contrast(sample, seed)
        sample = self.random_flipping(sample, seed)
        sample = self.random_cropping(sample, seed)
        if not is_mask:
            sample = self.random_rotate(sample, seed)
            # sample = tf.py_function(func=self.random_rotate, inp=[sample, seed], Tout=tf.float32)
        else:
            sample = self.random_rotate(sample, seed, 0)
            # sample = tf.py_function(func=self.random_rotate, inp=[sample, seed, 0], Tout=tf.uint8)

        return self.pad_to_original_size(sample)



import copy

from typing import Tuple
from typing import Union

import numpy as np
import SimpleITK as sitk
import PrognosAIs.Constants


class config:
    def __init__(self, config_settings: Union[dict, None]):
        if config_settings is not None:
            config_settings = copy.deepcopy(config_settings)
            self.perform_step = True
            for i_config_name, i_config_value in config_settings.items():
                setattr(self, i_config_name, i_config_value)

        else:
            self.perform_step = False

    @staticmethod
    def get_step_type(config: Union[dict, None]) -> Tuple[bool, bool, dict]:
        if config is not None:
            if "type" in config:
                step_type = config["type"]
                if step_type == "image":
                    perform_step_on_image = True
                    perform_step_on_patch = False
                elif step_type == "patch":
                    perform_step_on_image = False
                    perform_step_on_patch = True
                elif step_type == "both":
                    perform_step_on_image = True
                    perform_step_on_patch = True
                else:
                    raise NotImplementedError(
                        "You have specified an unknown step type {}!".format(step_type)
                    )
            else:
                perform_step_on_image = True
                perform_step_on_patch = False
        else:
            perform_step_on_image = perform_step_on_patch = False

        return perform_step_on_image, perform_step_on_patch, config


class general_config(config):
    def __init__(self, config_settings: dict):
        config_settings = copy.deepcopy(config_settings)
        general_config = config_settings.pop("general", None)
        self.pipeline = [
            "multi_dimension_extracting",
            "bias_field_correcting",
            "masking",
            "resampling",
            "normalizing",
            "rejecting",
            "patching",
            "saving",
        ]

        self.mask_keyword = "mask"
        self.max_cpus = 999
        super().__init__(general_config)


class multi_dimension_extracting_config(config):
    def __init__(self, config_settings: dict):
        self.extraction_type = None
        self.max_dimensions = -1
        self.perform_step_on_image = True
        self.perform_step_on_patch = False
        self.extract_masks = False

        super().__init__(config_settings)


class masking_config(config):
    def __init__(self, config_settings: dict):
        self.mask_background = False
        self.crop_to_mask = False
        self.background_value = 0.0
        self.process_masks = True
        self._mask_file = None
        self._mask = None

        (
            self.perform_step_on_image,
            self.perform_step_on_patch,
            config_settings,
        ) = super().get_step_type(config_settings)
        super().__init__(config_settings)

    @property
    def mask_file(self):
        return self._mask_file

    @mask_file.setter
    def mask_file(self, mask_file: str):
        self._mask_file = mask_file
        self.mask = self._mask_file

    @property
    def mask(self):
        return self._mask

    @mask.setter
    def mask(self, mask_file: str):
        mask = sitk.ReadImage(mask_file, sitk.sitkUInt8)
        mask = sitk.RescaleIntensity(mask, 0, 1)
        self._mask = mask


class resampling_config(config):
    def __init__(self, config_settings: dict):
        self.resample_size = [0, 0, 0]
        (
            self.perform_step_on_image,
            self.perform_step_on_patch,
            config_settings,
        ) = super().get_step_type(config_settings)

        super().__init__(config_settings)


class normalizing_config(config):
    def __init__(self, config_settings: dict):
        self._mask = None
        self._mask_file = None
        self.normalization_range = [0, 100]
        self.output_range = None
        self.mask_normalization = None
        self.normalization_method = None
        self.mask_smoothing = False

        (
            self.perform_step_on_image,
            self.perform_step_on_patch,
            config_settings,
        ) = super().get_step_type(config_settings)

        super().__init__(config_settings)

    @property
    def mask_file(self):
        return self._mask_file

    @mask_file.setter
    def mask_file(self, mask_file: str):
        self._mask_file = mask_file
        self.mask = self._mask_file

    @property
    def mask(self):
        return self._mask

    @mask.setter
    def mask(self, mask_file: str):
        mask = sitk.ReadImage(mask_file, sitk.sitkUInt8)
        mask = sitk.RescaleIntensity(mask, 0, 1)
        self._mask = mask


class bias_field_correcting_config(config):
    def __init__(self, config_settings: dict):
        self._mask_file = None
        self._mask = None

        (
            self.perform_step_on_image,
            self.perform_step_on_patch,
            config_settings,
        ) = super().get_step_type(config_settings)

        super().__init__(config_settings)

    @property
    def mask_file(self):
        return self._mask_file

    @mask_file.setter
    def mask_file(self, mask_file: str):
        self._mask_file = mask_file
        self.mask = self._mask_file

    @property
    def mask(self):
        return self._mask

    @mask.setter
    def mask(self, mask_file: str):
        mask = sitk.ReadImage(mask_file, sitk.sitkUInt8)
        mask = sitk.RescaleIntensity(mask, 0, 1)
        self._mask = mask


class patching_config(config):
    def __init__(self, config_settings: dict):
        self._patch_size = np.asarray([0, 0, 0])
        self.pad_if_needed = False
        self.pad_constant = 0.0
        self.extraction_type = "fitting"
        self.max_number_of_patches = -1
        self.overlap_fraction = 0.5

        self.perform_step_on_image = True
        self.perform_step_on_patch = False

        super().__init__(config_settings)

    @property
    def patch_size(self):
        return self._patch_size

    @patch_size.setter
    def patch_size(self, patch_size: list):
        self._patch_size = np.asarray(patch_size)


class rejecting_config(config):
    def __init__(self, config_settings: dict):
        self.rejection_limit = 0
        self.rejection_as_label = False
        self._mask_file = None
        self._mask = None
        (
            self.perform_step_on_image,
            self.perform_step_on_patch,
            config_settings,
        ) = super().get_step_type(config_settings)

        super().__init__(config_settings)

    @property
    def mask_file(self):
        return self._mask_file

    @mask_file.setter
    def mask_file(self, mask_file: str):
        self._mask_file = mask_file
        self.mask = self._mask_file

    @property
    def mask(self):
        return self._mask

    @mask.setter
    def mask(self, mask_file: str):
        mask = sitk.ReadImage(mask_file, sitk.sitkUInt8)
        mask = sitk.RescaleIntensity(mask, 0, 1)
        self._mask = mask


class saving_config(config):
    def __init__(self, config_settings: dict):
        self.use_mask_as_channel = False
        self.use_mask_as_label = False
        self.out_dir_name = "Samples"
        self.sample_npz_keyword = PrognosAIs.Constants.FEATURE_INDEX
        self.label_npz_keyword = PrognosAIs.Constants.LABEL_INDEX
        self.default_keyword = "default"
        self.named_channels = False
        self.impute_missing_channels = False
        self.save_as_float16 = False
        self.float16_percentage_diff = 0.1
        self.combine_labels = False
        self.channel_names = None
        self.mask_channels = 0
        (
            self.perform_step_on_image,
            self.perform_step_on_patch,
            config_settings,
        ) = super().get_step_type(config_settings)

        super().__init__(config_settings)
        self.perform_step = True
        if not self.perform_step_on_image and not self.perform_step_on_patch:
            self.perform_step_on_patch = True

class labeling_config(config):
    def __init__(self, config_settings: dict):
        config_settings = copy.deepcopy(config_settings)
        labeling_config = config_settings.pop("labeling", None)

        self.label_file = None
        self.train_fraction = 1.0
        self.validation_fraction = 0.0
        self.test_fraction = 0.0
        self.stratify_label_name = None
        self.filter_missing = False
        self.missing_value = -1
        self.make_one_hot = False

        super().__init__(labeling_config)



from PrognosAIs.Model.Architectures.Architecture import ClassificationNetworkArchitecture
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import Conv3D
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import Flatten
from tensorflow.keras.layers import MaxPooling2D
from tensorflow.keras.layers import MaxPooling3D
from tensorflow.keras.layers import PReLU
from tensorflow.keras.models import Model


class DDSNet(ClassificationNetworkArchitecture):
    def init_dimensionality(self, N_dimension):
        if N_dimension == 2:
            self.dims = 2
            self.conv_func = Conv2D
            self.pool_func = MaxPooling2D
        elif N_dimension == 3:
            self.dims = 3
            self.conv_func = Conv3D
            self.pool_func = MaxPooling3D

    def get_DDS_block(self, layer, N_filters):
        conv1 = self.conv_func(N_filters, [5] * self.dims)(layer)
        batch1 = BatchNormalization(axis=-1)(conv1)
        relu1 = PReLU()(batch1)
        conv2 = self.conv_func(N_filters, [5] * self.dims)(relu1)
        batch2 = BatchNormalization(axis=-1)(conv2)
        relu2 = PReLU()(batch2)
        pooling1 = self.pool_func(pool_size=[3] * self.dims)(relu2)

        return pooling1


class DDSNet_2D(DDSNet):
    dims = 2

    def create_model(self):
        self.check_minimum_input_size(self.inputs, [131, 131])
        self.init_dimensionality(self.dims)
        block_1 = self.get_DDS_block(self.inputs, 32)

        block_2 = self.get_DDS_block(block_1, 64)

        block_3 = self.get_DDS_block(block_2, 64)

        flattened_CNN_output = Flatten()(block_3)

        flattened_CNN_output = self.make_dropout_layer(flattened_CNN_output)

        dense_1 = Dense(1024, activation="relu")(flattened_CNN_output)

        dense_1 = self.make_dropout_layer(dense_1)

        predictions = self.outputs(dense_1)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model


class DDSNet_3D(DDSNet):
    dims = 3

    def create_model(self):
        self.check_minimum_input_size(self.inputs, [131, 131, 131])
        self.init_dimensionality(self.dims)
        block_1 = self.get_DDS_block(self.inputs, 32)

        block_2 = self.get_DDS_block(block_1, 64)

        block_3 = self.get_DDS_block(block_2, 64)

        flattened_CNN_output = Flatten()(block_3)

        flattened_CNN_output = self.make_dropout_layer(flattened_CNN_output)

        dense_1 = Dense(1024, activation="relu")(flattened_CNN_output)

        dense_1 = self.make_dropout_layer(dense_1)

        predictions = self.outputs(dense_1)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model



import numpy as np

from PrognosAIs.Model.Architectures.Architecture import NetworkArchitecture
from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import Concatenate
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import Conv3D
from tensorflow.keras.layers import Cropping2D
from tensorflow.keras.layers import Cropping3D
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import Dropout
from tensorflow.keras.layers import Flatten
from tensorflow.keras.layers import MaxPooling2D
from tensorflow.keras.layers import MaxPooling3D
from tensorflow.keras.layers import UpSampling2D
from tensorflow.keras.layers import UpSampling3D
from tensorflow.keras.layers import ZeroPadding2D
from tensorflow.keras.layers import ZeroPadding3D
from tensorflow.keras.models import Model
from tensorflow.keras import regularizers


class Unet(NetworkArchitecture):
    def make_outputs(
        self, output_info: dict, output_data_type: str, activation_type: str = "softmax"
    ):
        outputs = []
        if self.model_config is not None and "one_hot_output" in self.model_config:
            self.one_hot_output = self.model_config["one_hot_output"]
        else:
            self.one_hot_output = False

        for i_output_name, i_output_classes in output_info.items():
            if i_output_classes == 2 and not self.one_hot_output:
                temp_output = self.conv_func(
                    filters=1,
                    kernel_size=1,
                    padding="same",
                    dtype="float32",
                    activation="sigmoid",
                    name=i_output_name,
                )
            else:
                temp_output = self.conv_func(
                    filters=i_output_classes,
                    kernel_size=1,
                    padding="same",
                    dtype="float32",
                    activation="softmax",
                    name=i_output_name,
                )

            outputs.append(temp_output)
        if len(outputs) == 1:
            outputs = outputs[0]

        return outputs

    def get_number_of_filters(self):
        if self.model_config is not None and "number_of_filters" in self.model_config:
            return self.model_config["number_of_filters"]
        else:
            return 64

    def get_depth(self):
        if self.model_config is not None and "depth" in self.model_config:
            return self.model_config["depth"]
        else:
            return 5

    def init_dimensionality(self, N_dimension):
        if N_dimension == 2:
            self.dims = 2
            self.conv_func = Conv2D
            self.pool_func = MaxPooling2D
            self.upsample_func = UpSampling2D
            self.padding_func = ZeroPadding2D
            self.cropping_func = Cropping2D
        elif N_dimension == 3:
            self.dims = 3
            self.conv_func = Conv3D
            self.pool_func = MaxPooling3D
            self.upsample_func = UpSampling3D
            self.padding_func = ZeroPadding3D
            self.cropping_func = Cropping3D

    def get_conv_block(
        self, layer, N_filters, kernel_size=3, activation="relu", kernel_regularizer=None
    ):
        return self.conv_func(
            filters=N_filters,
            kernel_size=kernel_size,
            padding="same",
            activation=activation,
            kernel_regularizer=kernel_regularizer,
        )(layer)

    def get_pool_block(self, layer):
        return self.pool_func(pool_size=2, padding="same")(layer)

    def get_upsampling_block(self, layer, N_filters, activation="relu", kernel_regularizer=None):
        upsampling_layer = self.upsample_func(size=2)(layer)

        conv_layer = self.conv_func(
            filters=N_filters,
            kernel_size=2,
            padding="same",
            activation=activation,
            kernel_regularizer=kernel_regularizer,
        )(upsampling_layer)
        return conv_layer

    def get_padding_block(self, layer):
        input_size = layer.get_shape()
        input_size = np.asarray(input_size[1:-1])

        is_odd = []
        required_padding = []
        for i_input_size in input_size:
            if i_input_size % 2 != 0:
                is_odd.append(True)
                required_padding.append((1, 0))
            else:
                is_odd.append(False)
                required_padding.append((0, 0))

        if any(is_odd):
            required_padding = tuple(required_padding)
            padding_layer = self.padding_func(required_padding)(layer)
        else:
            padding_layer = layer
        return padding_layer

    def get_cropping_block(self, conv_layer, upsampling_layer):
        conv_size = conv_layer.get_shape()
        upsampling_size = upsampling_layer.get_shape()

        conv_size = np.asarray(conv_size[1:-1])
        upsampling_size = np.asarray(upsampling_size[1:-1])

        size_difference = upsampling_size - conv_size

        need_cropping = []
        required_crop = []
        for i_size_difference in size_difference:
            if i_size_difference != 0:
                need_cropping.append(True)
                required_crop.append((i_size_difference, 0))
            else:
                need_cropping.append(False)
                required_crop.append((0, 0))

        if any(need_cropping):
            required_crop = tuple(required_crop)
            cropping_layer = self.cropping_func(required_crop)(upsampling_layer)
        else:
            cropping_layer = upsampling_layer
        return cropping_layer


class UNet_2D(Unet):
    dims = 2

    def create_model(self):
        self.init_dimensionality(self.dims)
        self.inputs = self.make_inputs(self.input_shapes, self.input_data_type)
        self.N_filters = self.get_number_of_filters()
        self.depth = self.get_depth()

        head = self.inputs

        skip_layers = []
        for i_depth in range(self.depth - 1):
            head = self.get_conv_block(head, self.N_filters * (2 ** i_depth))
            head = self.get_conv_block(head, self.N_filters * (2 ** i_depth))
            head = self.make_dropout_layer(head)
            skip_layers.append(head)
            head = self.get_padding_block(head)
            head = self.get_pool_block(head)

        head = self.get_conv_block(head, self.N_filters * (2 ** self.depth))
        head = self.get_conv_block(head, self.N_filters * (2 ** self.depth))
        head = self.make_dropout_layer(head)

        for i_depth in range(self.depth - 2, -1, -1):
            head = self.get_upsampling_block(head, self.N_filters * (2 ** i_depth))
            head = self.get_cropping_block(skip_layers[i_depth], head)
            head = Concatenate(axis=-1)([skip_layers[i_depth], head])
            head = self.make_dropout_layer(head)

            head = self.get_conv_block(head, self.N_filters * (2 ** i_depth))
            head = self.get_conv_block(head, self.N_filters * (2 ** i_depth))

        outputs = self.make_outputs(self.output_info, self.output_data_type)
        predictions = outputs(head)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model


class UNet_3D(Unet):
    dims = 3

    def create_model(self):
        self.init_dimensionality(self.dims)
        self.inputs = self.make_inputs(self.input_shapes, self.input_data_type)
        self.N_filters = self.get_number_of_filters()
        self.depth = self.get_depth()

        head = self.inputs

        skip_layers = []
        for i_depth in range(self.depth - 1):
            head = self.get_conv_block(head, self.N_filters * (2 ** i_depth))
            head = self.get_conv_block(head, self.N_filters * (2 ** i_depth))
            head = self.make_dropout_layer(head)
            skip_layers.append(head)
            head = self.get_padding_block(head)
            head = self.get_pool_block(head)

        head = self.get_conv_block(head, self.N_filters * (2 ** self.depth))
        head = self.get_conv_block(head, self.N_filters * (2 ** self.depth))
        head = self.make_dropout_layer(head)

        for i_depth in range(self.depth - 2, -1, -1):
            head = self.get_upsampling_block(head, self.N_filters * (2 ** i_depth))
            head = self.get_cropping_block(skip_layers[i_depth], head)
            head = Concatenate(axis=-1)([skip_layers[i_depth], head])

            head = self.get_conv_block(head, self.N_filters * (2 ** i_depth))
            head = self.get_conv_block(head, self.N_filters * (2 ** i_depth))
            head = self.make_dropout_layer(head)

        outputs = self.make_outputs(self.output_info, self.output_data_type)
        predictions = outputs(head)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model



import time

import tensorflow as tf


def calculate_concordance_index(y_true, y_pred):
    """
    This function determine the concordance index for two numpy arrays

    y_true contains a label to indicate whether events occurred, and time to events
    (or time to right censored data if no event occurred)

    y_pred is beta*x in the cox model
    """
    events_occurred = y_true[:, 0]
    time_to_events = y_true[:, 1]

    total_c = 0.0
    N = 0.0
    for i_event_occurred, i_time_to_event, i_h in zip(events_occurred, time_to_events, y_pred):
        if i_event_occurred == 1:
            for j_time_to_event, j_h in zip(time_to_events, y_pred):
                if j_time_to_event > i_time_to_event:
                    if i_h < j_h:
                        total_c += 1.0
                    N += 1.0

    concordance_index = total_c / N

    return concordance_index


class ConcordanceIndex(tf.keras.callbacks.Callback):
    """
    A custom callback function to evaluate the concordance index on the
    whole validation set
    """

    def __init__(self, validation_generator):
        self.validation_generator = validation_generator
        self.validation_labels = validation_generator.get_labels()
        return

    def on_epoch_end(self, epoch, logs=None):
        y_pred = self.model.predict(self.validation_generator)
        y_real = self.validation_labels

        c_index = calculate_concordance_index(y_real, y_pred)

        logs["val_c_index"] = c_index
        return


class Timer(tf.keras.callbacks.Callback):
    """
    A custom callback function to evaluate the elapsed time of training
    """

    def __init__(self):
        self.start_time = time.time()
        return

    def on_epoch_end(self, epoch, logs=None):
        elapsed_time = time.time() - self.start_time

        logs["elapsed_time"] = elapsed_time
        return



import argparse
import copy
import os
import shutil
import sys
import warnings

import PrognosAIs.IO.utils as IO_utils

from PrognosAIs.IO import ConfigLoader
from PrognosAIs.IO import Configs
from PrognosAIs.Model import Evaluators
from PrognosAIs.Model import Trainer
from PrognosAIs.Preprocessing import Preprocessors
from slurmpie import slurmpie


class Pipeline(object):
    def __init__(
        self, config_file: str, preprocess: bool = True, train: bool = True, evaluate: bool = True, samples_folder: str = None,
    ):
        self.path = os.path.dirname(os.path.realpath(__file__))

        self.config_file = config_file
        self.config = ConfigLoader.ConfigLoader(config_file)
        self.input_folder = self.config.get_input_folder()
        self.output_folder = os.path.join(
            self.config.get_output_folder(), self.config.get_specific_output_folder()
        )
        if os.path.exists(self.output_folder):
            temp_i = 0
            original_folder = copy.deepcopy(self.output_folder)
            while os.path.exists(self.output_folder):
                self.output_folder = original_folder + "_" + str(temp_i)
                temp_i += 1
        IO_utils.create_directory(self.output_folder)
        self.config_file = self.config.copy_config(self.output_folder)

        # Paths to the python scripts we need
        self.preprocessor = os.path.join(self.path, "Preprocessing", "Preprocessors.py")
        self.trainer = os.path.join(self.path, "Model", "Trainer.py")
        self.evaluator = os.path.join(self.path, "Model", "Evaluators.py")

        self.preprocess = preprocess
        self.train = train
        self.evaluate = evaluate

        self.preprocessing_config = self.config.get_preprocessings_settings()

        if "saving" in self.preprocessing_config:
            saving_config = Configs.saving_config(self.preprocessing_config["saving"])
            print(saving_config)
        else:
            saving_config = Configs.saving_config(None)

        if samples_folder is None:
            self.samples_folder = os.path.join(self.output_folder, saving_config.out_dir_name)
        else:
            self.samples_folder = samples_folder

    def start_local_pipeline(self):

        if self.preprocess:
            preprocessor = Preprocessors.BatchPreprocessor(
                self.input_folder, self.output_folder, self.preprocessing_config
            )

            self.samples_folder = preprocessor.start()

        if self.train:
            trainer = Trainer.Trainer(self.config, self.samples_folder, self.output_folder)

            self.model_file = trainer.train_model()

        if self.train and self.evaluate:
            evaluator = Evaluators.Evaluator(
                self.model_file, self.samples_folder, self.config_file, self.output_folder,
            )
            evaluator.evaluate()

        elif self.evaluate and not self.train:
            warnings.warn(
                "Cannot evaluate if no model is trained, not doing evaluation", SyntaxWarning
            )

    def start_slurm_pipeline(
        self, preprocess_job: slurmpie.Job, train_job: slurmpie.Job, evaluate_job: slurmpie.Job
    ):
        save_name = self.config.get_save_name()
        self.model_file = os.path.join(self.output_folder, "MODEL", save_name + ".hdf5")
        preprocessing_command = """python -u {command} --config={config_file} --input={input_dir} --output={output_dir}""".format(
            command=self.preprocessor,
            config_file=self.config_file,
            input_dir=self.input_folder,
            output_dir=self.output_folder,
        )

        training_command = "srun python {command} --config={config_file} --input={input_dir} --output={output_dir} --savename={savename}".format(
            command=self.trainer,
            config_file=self.config_file,
            input_dir=self.samples_folder,
            output_dir=self.output_folder,
            savename=save_name,
        )

        evaluation_command = "python -u {command} --config={config_file} --input={input_dir} --output={output_dir} --model={model_file}".format(
            command=self.evaluator,
            config_file=self.config_file,
            input_dir=self.samples_folder,
            output_dir=self.output_folder,
            model_file=self.model_file,
        )

        # Here we make the actual jobs
        prognosAIs_pipeline = slurmpie.Pipeline()

        preprocess_job.script_is_file = False
        if preprocess_job.script is not None:
            preprocess_job.script += preprocessing_command
        else:
            preprocess_job.script = preprocessing_command

        train_job.script_is_file = False
        if train_job.script is not None:
            train_job.script += training_command
        else:
            train_job.script = training_command

        evaluate_job.script_is_file = False
        if evaluate_job.script is not None:
            evaluate_job.script += evaluation_command
        else:
            evaluate_job.script = evaluation_command

        if self.preprocess:
            prognosAIs_pipeline.add(preprocess_job)
        if self.preprocess and self.train:
            prognosAIs_pipeline.add({"afterok": [train_job]}, parent_job=preprocess_job)
        elif self.train:
            prognosAIs_pipeline.add(train_job)

        if self.evaluate:
            if self.train:
                prognosAIs_pipeline.add({"afterok": [evaluate_job]}, parent_job=train_job)
            elif self.preprocess:
                prognosAIs_pipeline.add({"afterok": [evaluate_job]}, parent_job=preprocess_job)
            else:
                prognosAIs_pipeline.add(evaluate_job)

        prognosAIs_pipeline.submit()
        print("Submitted the PrognosAIs pipeline!")

    @classmethod
    def init_from_sys_args(cls, args_in):
        parser = argparse.ArgumentParser(description="Start a PrognosAIs pipeline")
        parser.add_argument(
            "-c",
            "--config",
            required=True,
            help="The location of the PrognosAIs config file",
            metavar="configuration file",
            dest="config",
            type=str,
        )

        parser.add_argument(
            "-p",
            "--preprocess",
            required=False,
            help="Whether the pipeline should include preprocessing",
            dest="preprocess",
            action="store_true",
        )

        parser.add_argument(
            "-t",
            "--train",
            required=False,
            help="Whether the pipeline should include training",
            dest="train",
            action="store_true",
        )

        parser.add_argument(
            "-e",
            "--evaluate",
            required=False,
            help="Whether the pipeline should include evaluation",
            dest="evaluate",
            action="store_true",
        )

        args = parser.parse_args(args_in)

        if not args.preprocess and not args.train and not args.evaluate:
            args.preprocess = True
            args.train = True
            args.evaluate = True

        pipeline = cls(args.config, args.preprocess, args.train, args.evaluate)

        return pipeline


if __name__ == "__main__":
    pipeline = Pipeline.init_from_sys_args(sys.argv[1:])
    pipeline.start_local_pipeline()
